import struct
import datetime
from typing import Any
from array import array

import numpy as np
from numpy import dtype as np_dtype
import pandas as pd
from pandas.api import types as pd_types

from mindsdb.api.executor.sql_query.result_set import ResultSet, get_mysql_data_type_from_series, Column
from mindsdb.api.mysql.mysql_proxy.utilities.lightwood_dtype import dtype as lightwood_dtype
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    MYSQL_DATA_TYPE,
    DATA_C_TYPE_MAP,
    CTypeProperties,
    CHARSET_NUMBERS,
)
from mindsdb.utilities import log
from mindsdb.utilities.json_encoder import CustomJSONEncoder

logger = log.getLogger(__name__)

json_encoder = CustomJSONEncoder()


def column_to_mysql_column_dict(column: Column, database_name: str | None = None) -> dict[str, str | int]:
    """Convert Column object to dict with column properties.

    Args:
        column (Column): Column object to convert.
        database_name (str | None): Name of the database.

    Returns:
        dict[str, str | int]: Dictionary with mysql column properties.
    """
    # region infer type. Should not happen, but what if it is dtype of lightwood type?
    if isinstance(column.type, str):
        try:
            column.type = MYSQL_DATA_TYPE(column.type)
        except ValueError:
            if column.type == lightwood_dtype.date:
                column.type = MYSQL_DATA_TYPE.DATE
            elif column.type == lightwood_dtype.datetime:
                column.type = MYSQL_DATA_TYPE.DATETIME
            elif column.type == lightwood_dtype.float:
                column.type = MYSQL_DATA_TYPE.FLOAT
            elif column.type == lightwood_dtype.integer:
                column.type = MYSQL_DATA_TYPE.INT
            else:
                column.type = MYSQL_DATA_TYPE.TEXT
    elif isinstance(column.type, np_dtype):
        if pd_types.is_integer_dtype(column.type):
            column.type = MYSQL_DATA_TYPE.INT
        elif pd_types.is_numeric_dtype(column.type):
            column.type = MYSQL_DATA_TYPE.FLOAT
        elif pd_types.is_datetime64_any_dtype(column.type):
            column.type = MYSQL_DATA_TYPE.DATETIME
        else:
            column.type = MYSQL_DATA_TYPE.TEXT
    # endregion

    if isinstance(column.type, MYSQL_DATA_TYPE) is False:
        logger.warning(f"Unexpected column type: {column.type}. Use TEXT as fallback.")
        column.type = MYSQL_DATA_TYPE.TEXT

    charset = CHARSET_NUMBERS["utf8_unicode_ci"]
    if column.type in (MYSQL_DATA_TYPE.JSON, MYSQL_DATA_TYPE.VECTOR):
        charset = CHARSET_NUMBERS["binary"]

    type_properties: CTypeProperties = DATA_C_TYPE_MAP[column.type]

    result = {
        "database": column.database or database_name,
        #  TODO add 'original_table'
        "table_name": column.table_name,
        "name": column.name,
        "alias": column.alias or column.name,
        "size": type_properties.size,
        "flags": type_properties.flags,
        "type": type_properties.code,
        "charset": charset,
    }
    return result


def _dump_bool(var: Any) -> int | None:
    """Dumps a boolean value to an integer, as in MySQL boolean type is tinyint with values 0 and 1.
    NOTE: None consider as True in dataframe with dtype=bool, we can't change it

    Args:
        var (Any): The boolean value to dump

    Returns:
        int | None: 1 or 0 or None
    """
    if pd.isna(var):
        return None
    return "1" if var else "0"


def _dump_str(var: Any) -> str | None:
    """Dumps a value to a string.

    Args:
        var (Any): The value to dump

    Returns:
        str | None: The string representation of the value or None if the value is None
    """
    if isinstance(var, bytes):
        try:
            return var.decode("utf-8")
        except Exception:
            return str(var)[2:-1]
    if isinstance(var, (dict, list)):
        try:
            return json_encoder.encode(var)
        except Exception:
            return str(var)
    # pd.isna returns array of bools for list
    # and the truth value of a numpy array is ambiguous
    if isinstance(var, (list, np.ndarray)) is False and pd.isna(var):
        return None
    return str(var)


def _dump_int_or_str(var: Any) -> str | None:
    """Dumps a value to a string.
    If the value is numeric - then cast it to int to avoid float representation.

    Args:
        var (Any): The value to dump.

    Returns:
        str | None: The string representation of the value or None if the value is None
    """
    if pd.isna(var):
        return None
    try:
        return str(int(var))
    except ValueError:
        return str(var)


def _dump_date(var: datetime.date | str | None) -> str | None:
    """Dumps a date value to a string.

    Args:
        var (datetime.date | str | None): The date value to dump

    Returns:
        str | None: The string representation of the date value or None if the value is None
    """
    if isinstance(var, (datetime.date, pd.Timestamp)):  # it is also True for datetime.datetime
        return var.strftime("%Y-%m-%d")
    elif isinstance(var, str):
        return var
    elif pd.isna(var):
        return None
    logger.warning(f"Unexpected value type for DATE: {type(var)}, {var}")
    return _dump_str(var)


def _dump_datetime(var: datetime.datetime | str | None) -> str | None:
    """Dumps a datetime value to a string.
    # NOTE mysql may display only %Y-%m-%d %H:%M:%S format for datetime column

    Args:
        var (datetime.datetime | str | None): The datetime value to dump

    Returns:
        str | None: The string representation of the datetime value or None if the value is None
    """
    if isinstance(var, datetime.date):  # it is also datetime.datetime
        if hasattr(var, "tzinfo") and var.tzinfo is not None:
            return var.astimezone(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        return var.strftime("%Y-%m-%d %H:%M:%S")
    elif isinstance(var, pd.Timestamp):
        if var.tzinfo is not None:
            return var.tz_convert("UTC").strftime("%Y-%m-%d %H:%M:%S")
        return var.strftime("%Y-%m-%d %H:%M:%S")
    elif isinstance(var, str):
        return var
    elif pd.isna(var):
        return None
    logger.warning(f"Unexpected value type for DATETIME: {type(var)}, {var}")
    return _dump_str(var)


def _dump_time(var: datetime.time | str | None) -> str | None:
    """Dumps a time value to a string.

    Args:
        var (datetime.time | str | None): The time value to dump

    Returns:
        str | None: The string representation of the time value or None if the value is None
    """
    if isinstance(var, datetime.time):
        if var.tzinfo is not None:
            # NOTE strftime does not support timezone, so we need to convert to UTC
            offset_seconds = var.tzinfo.utcoffset(None).total_seconds()
            time_seconds = var.hour * 3600 + var.minute * 60 + var.second
            utc_seconds = (time_seconds - offset_seconds) % (24 * 3600)
            hours = int(utc_seconds // 3600)
            minutes = int((utc_seconds % 3600) // 60)
            seconds = int(utc_seconds % 60)
            var = datetime.time(hours, minutes, seconds, var.microsecond)
        return var.strftime("%H:%M:%S")
    elif isinstance(var, datetime.datetime):
        if var.tzinfo is not None:
            return var.astimezone(datetime.timezone.utc).strftime("%H:%M:%S")
        return var.strftime("%H:%M:%S")
    elif isinstance(var, pd.Timestamp):
        if var.tzinfo is not None:
            return var.tz_convert("UTC").strftime("%H:%M:%S")
        return var.strftime("%H:%M:%S")
    elif isinstance(var, str):
        return var
    elif pd.isna(var):
        return None
    logger.warning(f"Unexpected value type for TIME: {type(var)}, {var}")
    return _dump_str(var)


def _dump_vector(value: Any) -> bytes | None:
    """Convert array or list of floats to a bytes.

    Args:
        value (Any): The value to dump

    Returns:
        bytes | None: The bytes representation of the vector value or None if the value is None
    """
    if isinstance(value, (array, list, np.ndarray)):
        return b"".join([struct.pack("<f", el) for el in value])
    elif pd.isna(value):
        return None
    err_msg = f"Unexpected value type for VECTOR: {type(value)}, {value}"
    logger.error(err_msg)
    raise ValueError(err_msg)


def _handle_series_as_date(series: pd.Series) -> pd.Series:
    """Convert values in a series to a string representation of a date.
    NOTE: MySQL require exactly %Y-%m-%d for DATE type.

    Args:
        series (pd.Series): The series to handle

    Returns:
        pd.Series: The series with the date values as strings
    """
    if pd_types.is_datetime64_any_dtype(series.dtype):
        return series.dt.strftime("%Y-%m-%d")
    elif pd_types.is_object_dtype(series.dtype):
        return series.apply(_dump_date)
    logger.info(f"Unexpected dtype: {series.dtype} for column with type DATE")
    return series.apply(_dump_str)


def _handle_series_as_datetime(series: pd.Series) -> pd.Series:
    """Convert values in a series to a string representation of a datetime.
    NOTE: MySQL's DATETIME type require exactly %Y-%m-%d %H:%M:%S format.

    Args:
        series (pd.Series): The series to handle

    Returns:
        pd.Series: The series with the datetime values as strings
    """
    if pd_types.is_datetime64_any_dtype(series.dtype):
        return series.dt.strftime("%Y-%m-%d %H:%M:%S")
    elif pd_types.is_object_dtype(series.dtype):
        return series.apply(_dump_datetime)
    logger.info(f"Unexpected dtype: {series.dtype} for column with type DATETIME")
    return series.apply(_dump_str)


def _handle_series_as_time(series: pd.Series) -> pd.Series:
    """Convert values in a series to a string representation of a time.
    NOTE: MySQL's TIME type require exactly %H:%M:%S format.

    Args:
        series (pd.Series): The series to handle

    Returns:
        pd.Series: The series with the time values as strings
    """
    if pd_types.is_timedelta64_ns_dtype(series.dtype):
        base_time = pd.Timestamp("2000-01-01")
        series = (base_time + series).dt.strftime("%H:%M:%S")
    elif pd_types.is_datetime64_dtype(series.dtype):
        series = series.dt.strftime("%H:%M:%S")
    elif pd_types.is_object_dtype(series.dtype):
        series = series.apply(_dump_time)
    else:
        logger.info(f"Unexpected dtype: {series.dtype} for column with type TIME")
        series = series.apply(_dump_str)
    return series


def _handle_series_as_int(series: pd.Series) -> pd.Series:
    """Dump series to str(int) (or just str, of can't case to int). This need because of DataFrame store imput int as
    float if dtype is object: pd.DataFrame([None, 1], dtype='object') -> [NaN, 1.0]

    Args:
        series (pd.Series): The series to handle

    Returns:
        pd.Series: The series with the int values as strings
    """
    if pd_types.is_integer_dtype(series.dtype):
        if series.dtype == "Int64":
            # NOTE: 'apply' converts values to python floats
            return series.astype(object).apply(_dump_str)
        return series.apply(_dump_str)
    return series.apply(_dump_int_or_str)


def _handle_series_as_vector(series: pd.Series) -> pd.Series:
    """Convert values in a series to a bytes representation of a vector.
    NOTE: MySQL's VECTOR type require exactly 4 bytes per float.

    Args:
        series (pd.Series): The series to handle

    Returns:
        pd.Series: The series with the vector values as bytes
    """
    return series.apply(_dump_vector)


def dump_result_set_to_mysql(
    result_set: ResultSet, infer_column_size: bool = False
) -> tuple[pd.DataFrame, list[dict[str, str | int]]]:
    """
    Dumps the ResultSet to a format that can be used to send as MySQL response packet.
    NOTE: This method modifies the original DataFrame and columns.

    Args:
        result_set (ResultSet): result set to dump
        infer_column_size (bool): If True, infer the 'size' attribute of the column from the data.
                                  Exact size is not necessary, approximate is enough.

    Returns:
        tuple[pd.DataFrame, list[dict[str, str | int]]]: A tuple containing the modified DataFrame and a list
                                                            of MySQL column dictionaries. The dataframe values are
                                                            str or None, dtype=object
    """
    df = result_set.get_raw_df()

    for i, column in enumerate(result_set.columns):
        series = df[i]
        if isinstance(column.type, MYSQL_DATA_TYPE) is False:
            column.type = get_mysql_data_type_from_series(series)

        column_type: MYSQL_DATA_TYPE = column.type

        match column_type:
            case MYSQL_DATA_TYPE.BOOL | MYSQL_DATA_TYPE.BOOLEAN:
                series = series.apply(_dump_bool)
            case MYSQL_DATA_TYPE.DATE:
                series = _handle_series_as_date(series)
            case MYSQL_DATA_TYPE.DATETIME:
                series = _handle_series_as_datetime(series)
            case MYSQL_DATA_TYPE.TIME:
                series = _handle_series_as_time(series)
            case (
                MYSQL_DATA_TYPE.INT
                | MYSQL_DATA_TYPE.TINYINT
                | MYSQL_DATA_TYPE.SMALLINT
                | MYSQL_DATA_TYPE.MEDIUMINT
                | MYSQL_DATA_TYPE.BIGINT
                | MYSQL_DATA_TYPE.YEAR
            ):
                series = _handle_series_as_int(series)
            case MYSQL_DATA_TYPE.VECTOR:
                series = _handle_series_as_vector(series)
            case _:
                series = series.apply(_dump_str)

        # inplace modification of dt types raise SettingWithCopyWarning, so do regular replace
        # we may split this operation for dt and other types for optimisation
        df[i] = series.replace([np.NaN, pd.NA, pd.NaT], None)

    columns_dicts = [column_to_mysql_column_dict(column) for column in result_set.columns]

    if infer_column_size and any(column_info.get("size") is None for column_info in columns_dicts):
        if len(df) == 0:
            for column_info in columns_dicts:
                if column_info["size"] is None:
                    column_info["size"] = 1
        else:
            sample = df.head(100)
            for i, column_info in enumerate(columns_dicts):
                try:
                    column_info["size"] = sample[sample.columns[i]].astype(str).str.len().max()
                except Exception:
                    column_info["size"] = 1

    return df, columns_dicts



"""file_metadata

Revision ID: 4943359e354a
Revises: c06c35f7e8e1
Create Date: 2025-02-09 10:10:55.577407

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '4943359e354a'
down_revision = 'c06c35f7e8e1'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.add_column(sa.Column('metadata', sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.drop_column('metadata')
    # ### end Alembic commands ###



from typing import List
from textwrap import dedent
from datetime import datetime

from langchain_community.agent_toolkits.sql.toolkit import SQLDatabaseToolkit
from langchain_community.tools import ListSQLDatabaseTool, InfoSQLDatabaseTool, QuerySQLDataBaseTool
from langchain_core.tools import BaseTool

from mindsdb.interfaces.skills.custom.text2sql.mindsdb_sql_tool import MindsDBSQLParserTool
from mindsdb.interfaces.skills.custom.text2sql.mindsdb_kb_tools import (
    KnowledgeBaseListTool,
    KnowledgeBaseInfoTool,
    KnowledgeBaseQueryTool,
)


class MindsDBSQLToolkit(SQLDatabaseToolkit):
    include_tables_tools: bool = True
    include_knowledge_base_tools: bool = True

    def get_tools(self, prefix="") -> List[BaseTool]:
        current_date_time = datetime.now().strftime("%Y-%m-%d %H:%M")

        """Get the tools in the toolkit."""
        list_sql_database_tool = ListSQLDatabaseTool(
            name=f"sql_db_list_tables{prefix}",
            db=self.db,
            description=dedent(
                """\n
                Input is an empty string, output is a comma-separated list of tables in the database. Each table name is escaped using backticks.
                Each table name in the list may be in one of two formats: database_name.`table_name` or database_name.schema_name.`table_name`.
                Table names in response to the user must be escaped using backticks.
            """
            ),
        )

        info_sql_database_tool_description = (
            "Input: A comma-separated list of tables enclosed between the symbols $START$ and $STOP$. The tables names itself must be escaped using backticks.\n"
            "Output: Schema and sample rows for those tables. \n"
            "Use this tool to investigate table schemas for needed columns. "
            f"Ensure tables exist by calling {list_sql_database_tool.name} first. "
            # "The names of tables, schemas, and databases must be escaped using backticks. "
            # "Always enclose the names of tables, schemas, and databases in backticks. "
            "Get sample data with 'SELECT * FROM `database`.`table` LIMIT 3' before answering questions. \n"
            "Example of correct Input:\n    $START$ `database`.`table1`, `database`.`table2`, `database`.`table3` $STOP$\n"
            "    $START$ `table1` `table2` `table3` $STOP$\n"
            "Example of wrong Input:\n    $START$ `database.table1`, `database.table2`, `database.table3` $STOP$\n"
            "    $START$ table1 table2 table3 $STOP$\n"
        )
        info_sql_database_tool = InfoSQLDatabaseTool(
            name=f"sql_db_schema{prefix}", db=self.db, description=info_sql_database_tool_description
        )

        query_sql_database_tool_description = dedent(
            f"""\
            Input: A detailed and well-structured SQL query. The query must be enclosed between the symbols $START$ and $STOP$.
            Output: Database result or error message. For errors, rewrite and retry the query. For 'Unknown column' errors, use '{info_sql_database_tool.name}' to check table fields.
            This system is a highly intelligent and reliable PostgreSQL SQL skill designed to work with databases.
            Follow these instructions with utmost precision:
            1. Final Response Format:
               - Assume the frontend fully supports Markdown unless the user specifies otherwise.
               - When the response contains data that fits a table format, present it as a properly formatted Markdown table
               - Ensure clarity and proper structure for easy readability.
            2. Sample Data:
               - Before answering a question, if you don't have sample data about a table, **always** get sample data using `SELECT * FROM table LIMIT 3` from the tables you believe are relevant to formulating your answers.
            3. Categorical Data:
               - Whenever working with a column where values seem categorical, especially when filtering with `WHERE col = 'value'`, `WHERE col IN (list of values)`, or `WHERE col NOT IN (list of values)`, **always** retrieve the distinct values first.
               - Before writing your main query, always run `SELECT DISTINCT col` to fetch a list of unique values from that column. This step is mandatory to ensure accurate queries and responses.
            4. Result Limiting and Counting:
               - Unless instructed otherwise by the user, always run a count on the final query first using `SELECT COUNT(*)`.
               - If the count is greater than 10, limit the query to return only 10 results initially.
               - **Always** inform the user of the total number of results available and specify that you are providing the first 10 results.
               - Let the user know they can request additional results and/or specify how they would like the results ordered or grouped.
            5. Date Handling:
               - **System current date and time: {current_date_time} (UTC or local timezone based on server settings).**
               - **Always** use PostgreSQL-compatible `CURRENT_DATE` or `NOW()` functions when working with dates—never assume or guess the current date.
               - For any date-related comparisons in the query, *always* ensure that your query casts the column being compared using `column_name::DATE [operator] ..`
               - Do not compare date values without casting columns to date.
               - For date interval operations, use Interval units as keywords. You can use keywords to specify units like days, hours, months, years, etc., directly without quotes. Examples:
                 SELECT NOW() + INTERVAL 5 DAY;
                 SELECT NOW() - INTERVAL 3 HOUR;
                 SELECT NOW() + INTERVAL 2 MONTH + INTERVAL 3 DAY;
                 SELECT NOW() - INTERVAL 1 YEAR;
               - Always run SELECT NOW() to retrieve the current date when answering current or relative to current date-related questions.
            6. Query Best Practices:
               - Always send only one query at a time.
               - Always enclose the names of tables, schemas, and databases in backticks.
               - The input SQL query must end with a semicolon.
               - Query only necessary columns, not all.
               - Use only existing column names from correct tables.
               - Use database-specific syntax for date operations.
            7. Error Handling:
               - For errors, rewrite and retry the query.
               - For 'Unknown column' errors, check table fields using info_sql_database_tool.
            8. Identity and Purpose:
               - When asked about yourself or your maker, state that you are a Data-Mind, created by MindsDB to help answer data questions.
               - When asked about your purpose or how you can help, explore the available data sources and then explain that you can answer questions based on the connected data. Provide a few relevant example questions that you could answer for the user about their data.
            Adhere to these guidelines for all queries and responses. Ask for clarification if needed.
        """
        )

        query_sql_database_tool = QuerySQLDataBaseTool(
            name=f"sql_db_query{prefix}", db=self.db, description=query_sql_database_tool_description
        )

        mindsdb_sql_parser_tool_description = (
            "Use this tool to ensure a SQL query passes the MindsDB SQL parser. "
            "If the query is not correct, it will be corrected and returned. Use the new query. "
            "If the query can't be corrected, an error is returned. In this case, rewrite and retry. "
            "If the query is correct, it will be parsed and returned. "
            f"ALWAYS run this tool before executing a query with {query_sql_database_tool.name}. "
        )
        mindsdb_sql_parser_tool = MindsDBSQLParserTool(
            name=f"mindsdb_sql_parser_tool{prefix}", description=mindsdb_sql_parser_tool_description
        )

        sql_tools = [
            query_sql_database_tool,
            info_sql_database_tool,
            list_sql_database_tool,
            mindsdb_sql_parser_tool,
        ]
        if not self.include_knowledge_base_tools:
            return sql_tools

        # Knowledge base tools
        kb_list_tool = KnowledgeBaseListTool(
            name=f"kb_list_tool{prefix}",
            db=self.db,
            description=dedent(
                """\
                Lists all available knowledge bases that can be queried.
                Input: No input required, just call the tool directly.
                Output: A table of all available knowledge bases with their names and creation dates.

                Use this tool first when answering factual questions to see what knowledge bases are available.
                Each knowledge base name is escaped using backticks.

                Example usage: kb_list_tool()
            """
            ),
        )

        kb_info_tool = KnowledgeBaseInfoTool(
            name=f"kb_info_tool{prefix}",
            db=self.db,
            description=dedent(
                f"""\
                Gets detailed information about specific knowledge bases including their structure and metadata fields.

                Input: A knowledge base name as a simple string.
                Output: Schema, metadata columns, and sample rows for the specified knowledge base.

                Use this after kb_list_tool to understand what information is contained in the knowledge base
                and what metadata fields are available for filtering.

                Example usage: kb_info_tool("kb_name")

                Make sure the knowledge base exists by calling {kb_list_tool.name} first.
            """
            ),
        )

        kb_query_tool = KnowledgeBaseQueryTool(
            name=f"kb_query_tool{prefix}",
            db=self.db,
            description=dedent(
                f"""\
                Queries knowledge bases using SQL syntax to retrieve relevant information.

                Input: A SQL query string that targets a knowledge base.
                Output: Knowledge base search results or error message.

                This tool is designed for semantic search and metadata filtering in MindsDB knowledge bases.

                Query Types and Examples:
                1. Basic semantic search:
                   kb_query_tool("SELECT * FROM kb_name WHERE content = 'your search query';")

                2. Metadata filtering:
                   kb_query_tool("SELECT * FROM kb_name WHERE metadata_field = 'value';")

                3. Combined search:
                   kb_query_tool("SELECT * FROM kb_name WHERE content = 'query' AND metadata_field = 'value';")

                4. Setting relevance threshold:
                   kb_query_tool("SELECT * FROM kb_name WHERE content = 'query' AND relevance_threshold = 0.7;")

                5. Limiting results:
                   kb_query_tool("SELECT * FROM kb_name WHERE content = 'query' LIMIT 5;")

                6. Getting sample data:
                   kb_query_tool("SELECT * FROM kb_name LIMIT 3;")

                7. Don't use LIKE operator on content filter ie semantic search:
                SELECT * FROM `test_kb` WHERE content LIKE '%population of New York%' $STOP$

                Like is not supported, use the following instead:
                SELECT * FROM `test_kb` WHERE content = 'population of New York'

                Result Format:
                - Results include: id, chunk_id, chunk_content, metadata, distance, and relevance columns
                - The metadata column contains a JSON object with all metadata fields

                Best Practices:
                - Always check available knowledge bases with {kb_list_tool.name} first
                - Use {kb_info_tool.name} to understand the structure and metadata fields
                - Always include a semicolon at the end of your SQL query

                For factual questions, use this tool to retrieve information rather than relying on the model's knowledge.
            """
            ),
        )

        # Return standard SQL tools and knowledge base tools
        kb_tools = [
            kb_list_tool,
            kb_info_tool,
            kb_query_tool,
        ]

        if not self.include_tables_tools:
            return kb_tools
        else:
            return sql_tools + kb_tools



from typing import List

import pandas as pd

from mindsdb.integrations.handlers.confluence_handler.confluence_api_client import ConfluenceAPIClient
from mindsdb.integrations.libs.api_handler import APIResource
from mindsdb.integrations.utilities.sql_utils import (
    FilterCondition,
    FilterOperator,
    SortColumn
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class ConfluenceSpacesTable(APIResource):
    """
    The table abstraction for the 'spaces' resource of the Confluence API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'spaces' resource of the Confluence API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        spaces = []
        client: ConfluenceAPIClient = self.handler.connect()

        ids, keys, space_type, status = None, None, None, None
        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

            if condition.column == "key":
                if condition.op == FilterOperator.EQUAL:
                    keys = [condition.value]

                elif condition.op == FilterOperator.IN:
                    keys = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'key'."
                    )

                condition.applied = True

            if condition.column == "type":
                if condition.op == FilterOperator.EQUAL:
                    space_type = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'type'."
                    )

                condition.applied = True

            if condition.column == "status":
                if condition.op == FilterOperator.EQUAL:
                    status = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'status'."
                    )

                condition.applied = True

        sort_condition = None
        if sort:
            for sort_column in sort:
                if sort_column.column in ["id", "key", "name"]:
                    if sort_column.ascending:
                        sort_condition = sort_column.column

                    else:
                        sort_condition = f"-{sort_column.column}"

                    sort_column.applied = True
                    break

        spaces = client.get_spaces(
            ids=ids,
            keys=keys,
            space_type=space_type,
            status=status,
            sort_condition=sort_condition,
            limit=limit
        )

        spaces_df = pd.json_normalize(spaces, sep="_")
        spaces_df = spaces_df[self.get_columns()]

        return spaces_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'spaces' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'spaces' resource.
        """
        return [
            "id",
            "key",
            "name",
            "type",
            "description_view_representation",
            "description_view_value",
            "status",
            "authorId",
            "createdAt",
            "homepageId",
            "_links_webui",
            "currentActiveAlias",
        ]


class ConfluencePagesTable(APIResource):
    """
    The table abstraction for the 'pages' resource of the Confluence API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'pages' resource of the Confluence API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        pages = []
        client: ConfluenceAPIClient = self.handler.connect()

        page_ids, space_ids, statuses, title = None, None, None, None
        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    page_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    page_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'page_id'."
                    )

                condition.applied = True

            if condition.column == "spaceId":
                if condition.op == FilterOperator.EQUAL:
                    space_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    space_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'spaceId'."
                    )

                condition.applied = True

            if condition.column == "status":
                if condition.op == FilterOperator.EQUAL:
                    statuses = [condition.value]

                elif condition.op == FilterOperator.IN:
                    statuses = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'status'."
                    )

                condition.applied = True

            if condition.column == "title":
                if condition.op == FilterOperator.EQUAL:
                    title = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'title'."
                    )

                condition.applied = True

        sort_condition = None
        if sort:
            for sort_column in sort:
                if sort_column.column in ["id", "title", "createdAt"]:
                    sort_condition = sort_column.column if sort_column != "createdAt" else "created-date"
                    if not sort_column.ascending:
                        sort_condition = f"-{sort_condition}"

                    sort_column.applied = True
                    break

        pages = client.get_pages(
            page_ids=page_ids,
            space_ids=space_ids,
            statuses=statuses,
            title=title,
            sort_condition=sort_condition,
            limit=limit
        )

        pages_df = pd.json_normalize(pages, sep="_")
        pages_df = pages_df[self.get_columns()]

        return pages_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'pages' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'pages' resource.
        """
        return [
            "id",
            "status",
            "title",
            "spaceId",
            "parentId",
            "parentType",
            "position",
            "authorId",
            "ownerId",
            "lastOwnerId",
            "createdAt",
            "version_createdAt",
            "version_message",
            "version_number",
            "version_minorEdit",
            "version_authorId",
            "body_storage_representation",
            "body_storage_value",
            "_links_webui",
            "_links_editui",
            "_links_tinyui",
        ]


class ConfluenceBlogPostsTable(APIResource):
    """
    The table abstraction for the 'blogposts' resource of the Confluence API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'blogposts' resource of the Confluence API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        blogposts = []
        client: ConfluenceAPIClient = self.handler.connect()

        post_ids, space_ids, statuses, title = None, None, None, None
        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    post_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    post_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

            if condition.column == "spaceId":
                if condition.op == FilterOperator.EQUAL:
                    space_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    space_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'spaceKey'."
                    )

                condition.applied = True

            if condition.column == "status":
                if condition.op == FilterOperator.EQUAL:
                    statuses = [condition.value]

                elif condition.op == FilterOperator.IN:
                    statuses = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'status'."
                    )

                condition.applied = True

            if condition.column == "title":
                if condition.op == FilterOperator.EQUAL:
                    title = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'title'."
                    )

                condition.applied = True

        sort_condition = None
        if sort:
            for sort_column in sort:
                if sort_column.column in ["id", "title", "createdAt"]:
                    sort_condition = sort_column.column if sort_column != "createdAt" else "created-date"
                    if not sort_column.ascending:
                        sort_condition = f"-{sort_condition}"

                    sort_column.applied = True
                    break

        blogposts = client.get_blogposts(
            post_ids=post_ids,
            space_ids=space_ids,
            statuses=statuses,
            title=title,
            sort_condition=sort_condition,
            limit=limit
        )

        blogposts_df = pd.json_normalize(blogposts, sep="_")
        blogposts_df = blogposts_df[self.get_columns()]

        return blogposts_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'blogposts' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'blogposts' resource.
        """
        return [
            "id",
            "status",
            "title",
            "spaceId",
            "authorId",
            "createdAt",
            "version_createdAt",
            "version_message",
            "version_number",
            "version_minorEdit",
            "version_authorId",
            "body_storage_representation",
            "body_storage_value",
            "_links_webui",
            "_links_editui",
            "_links_tinyui",
        ]


class ConfluenceWhiteboardsTable(APIResource):
    """
    The table abstraction for the 'whiteboards' resource of the Confluence API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'whiteboards' resource of the Confluence API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        whiteboards = []
        client: ConfluenceAPIClient = self.handler.connect()

        whiteboard_ids = None
        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    whiteboard_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    whiteboard_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

        if not whiteboard_ids:
            raise ValueError("The 'id' column must be provided in the WHERE clause.")

        whiteboards = [client.get_whiteboard_by_id(whiteboard_id) for whiteboard_id in whiteboard_ids]

        whiteboards_df = pd.json_normalize(whiteboards, sep="_")
        whiteboards_df = whiteboards_df[self.get_columns()]

        return whiteboards_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'whiteboards' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'whiteboards' resource.
        """
        return [
            "id",
            "type",
            "status",
            "title",
            "parentId",
            "parentType",
            "position",
            "authorId",
            "ownerId",
            "createdAt",
            "version_createdAt",
            "version_message",
            "version_number",
            "version_minorEdit",
            "version_authorId",
        ]


class ConfluenceDatabasesTable(APIResource):
    """
    The table abstraction for the 'databases' resource of the Confluence API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'databases' resource of the Confluence API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        databases = []
        client: ConfluenceAPIClient = self.handler.connect()

        database_ids = None
        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    database_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    database_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

        if not database_ids:
            raise ValueError("The 'id' column must be provided in the WHERE clause.")

        databases = [client.get_database_by_id(database_id) for database_id in database_ids]

        databases_df = pd.json_normalize(databases, sep="_")
        databases_df = databases_df[self.get_columns()]

        return databases_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'databases' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'databases' resource.
        """
        return [
            "id",
            "type",
            "status",
            "title",
            "parentId",
            "parentType",
            "position",
            "authorId",
            "ownerId",
            "createdAt",
            "version_createdAt",
            "version_message",
            "version_number",
            "version_minorEdit",
            "version_authorId",
        ]


class ConfluenceTasksTable(APIResource):
    """
    The table abstraction for the 'tasks' resource of the Confluence API.
    """

    def list(
        self,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        sort: List[SortColumn] = None,
        targets: List[str] = None,
        **kwargs
    ):
        """
        Executes a parsed SELECT SQL query on the 'tasks' resource of the Confluence API.

        Args:
            conditions (List[FilterCondition]): The list of parsed filter conditions.
            limit (int): The maximum number of records to return.
            sort (List[SortColumn]): The list of parsed sort columns.
            targets (List[str]): The list of target columns to return.
        """
        tasks = []
        client: ConfluenceAPIClient = self.handler.connect()

        task_ids = None
        space_ids = None
        page_ids = None
        blogpost_ids = None
        created_by_ids = None
        assigned_to_ids = None
        completed_by_ids = None
        status = None

        for condition in conditions:
            if condition.column == "id":
                if condition.op == FilterOperator.EQUAL:
                    task_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    task_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'id'."
                    )

                condition.applied = True

            if condition.column == "spaceId":
                if condition.op == FilterOperator.EQUAL:
                    space_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    space_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'spaceId'."
                    )

                condition.applied = True

            if condition.column == "pageId":
                if condition.op == FilterOperator.EQUAL:
                    page_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    page_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'pageId'."
                    )

                condition.applied = True

            if condition.column == "blogPostId":
                if condition.op == FilterOperator.EQUAL:
                    blogpost_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    blogpost_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'blogPostId'."
                    )

                condition.applied = True

            if condition.column == "createdBy":
                if condition.op == FilterOperator.EQUAL:
                    created_by_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    created_by_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'createdBy'."
                    )

                condition.applied = True

            if condition.column == "assignedTo":
                if condition.op == FilterOperator.EQUAL:
                    assigned_to_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    assigned_to_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'assignedTo'."
                    )

                condition.applied = True

            if condition.column == "completedBy":
                if condition.op == FilterOperator.EQUAL:
                    completed_by_ids = [condition.value]

                elif condition.op == FilterOperator.IN:
                    completed_by_ids = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'completedBy'."
                    )

                condition.applied = True

            if condition.column == "status":
                if condition.op == FilterOperator.EQUAL:
                    status = condition.value

                else:
                    raise ValueError(
                        f"Unsupported operator '{condition.op}' for column 'status'."
                    )

                condition.applied = True

        tasks = client.get_tasks(
            task_ids=task_ids,
            space_ids=space_ids,
            page_ids=page_ids,
            blogpost_ids=blogpost_ids,
            created_by_ids=created_by_ids,
            assigned_to_ids=assigned_to_ids,
            completed_by_ids=completed_by_ids,
            status=status,
            limit=limit
        )
        tasks_df = pd.json_normalize(tasks, sep="_")

        # Each task will have either a 'pageId' or 'blogPostId' but not both.
        # In situations where they are all from the same resource, the other column will be empty.
        # We will fill the empty column with None to ensure consistency.
        for column in ["pageId", "blogPostId"]:
            if column not in tasks_df.columns:
                tasks_df[column] = None

        tasks_df = tasks_df[self.get_columns()]

        return tasks_df

    def get_columns(self) -> List[str]:
        """
        Retrieves the attributes (columns) of the 'tasks' resource.

        Returns:
            List[Text]: A list of attributes (columns) of the 'tasks' resource.
        """
        return [
            "id",
            "localId",
            "spaceId",
            "pageId",
            "blogPostId",
            "status",
            "body_storage_representation",
            "body_storage_value",
            "createdBy",
            "assignedTo",
            "completedBy",
            "createdAt",
            "updatedAt",
            "dueAt",
            "completedAt",
        ]



"""add agent_skills_table

Revision ID: e187961e844a
Revises: 4c26ad04eeaa
Create Date: 2023-09-06 13:56:17.803484

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e187961e844a'
down_revision = '4c26ad04eeaa'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'agent_skills',
        sa.Column('agent_id', sa.Integer(), nullable=False),
        sa.Column('skill_id', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('agent_id', 'skill_id')
    )


def downgrade():
    op.drop_table('agent_skills')



from typing import Optional, Dict, List, Tuple
import json
import itertools
import re

import numpy as np
import pandas as pd

from mindsdb.integrations.libs.llm.config import (
    AnthropicConfig,
    BaseLLMConfig,
    GoogleConfig,
    LiteLLMConfig,
    OllamaConfig,
    OpenAIConfig,
    NvidiaNIMConfig,
    MindsdbConfig,
    WriterConfig,
)
from mindsdb.utilities.config import config
from langchain_text_splitters import Language, RecursiveCharacterTextSplitter


# Default to latest GPT-4 model (https://platform.openai.com/docs/models/gpt-4-and-gpt-4-turbo)
DEFAULT_OPENAI_MODEL = "gpt-4o"
# Requires more than vanilla OpenAI due to ongoing summarization and 3rd party input.
DEFAULT_OPENAI_MAX_TOKENS = 8096
DEFAULT_OPENAI_MAX_RETRIES = 3

DEFAULT_ANTHROPIC_MODEL = "claude-3-haiku-20240307"

DEFAULT_GOOGLE_MODEL = "gemini-2.5-pro-preview-03-25"

DEFAULT_LITELLM_MODEL = "gpt-3.5-turbo"
DEFAULT_LITELLM_PROVIDER = "openai"
DEFAULT_LITELLM_BASE_URL = "https://ai.dev.mindsdb.com"

DEFAULT_OLLAMA_BASE_URL = "http://localhost:11434"
DEFAULT_OLLAMA_MODEL = "llama2"

DEFAULT_NVIDIA_NIM_BASE_URL = "http://localhost:8000/v1"  # Assumes local port forwarding through ssh
DEFAULT_NVIDIA_NIM_MODEL = "meta/llama-3_1-8b-instruct"
DEFAULT_VLLM_SERVER_URL = "http://localhost:8000/v1"


def get_completed_prompts(base_template: str, df: pd.DataFrame, strict=True) -> Tuple[List[str], np.ndarray]:
    """
    Helper method that produces formatted prompts given a template and data in a Pandas DataFrame.
    It also returns the ID of any empty templates that failed to be filled due to missing data.

    :param base_template: string with placeholders for each column in the DataFrame. Placeholders should follow double curly braces format, e.g. `{{column_name}}`. All placeholders should have matching columns in `df`.
    :param df: pd.DataFrame to generate full prompts. Each placeholder in `base_template` must exist as a column in the DataFrame. If a column is not in the template, it is ignored entirely.
    :param strict: raise exception if base_template doesn't contain placeholders

    :return prompts: list of in-filled prompts using `base_template` and relevant columns from `df`
    :return empty_prompt_ids: np.int numpy array (shape (n_missing_rows,)) with the row indexes where in-fill failed due to missing data.
    """  # noqa
    columns = []
    spans = []
    matches = list(re.finditer("{{(.*?)}}", base_template))

    if len(matches) == 0:
        # no placeholders
        if strict:
            raise AssertionError("No placeholders found in the prompt, please provide a valid prompt template.")
        prompts = [base_template] * len(df)
        return prompts, np.ndarray(0)

    first_span = matches[0].start()
    last_span = matches[-1].end()

    for m in matches:
        columns.append(m[0].replace("{", "").replace("}", ""))
        spans.extend((m.start(), m.end()))

    spans = spans[1:-1]  # omit first and last, they are added separately
    template = [
        base_template[s:e] for s, e in list(zip(spans, spans[1:]))[::2]
    ]  # take every other to skip placeholders  # noqa
    template.insert(0, base_template[0:first_span])  # add prompt start
    template.append(base_template[last_span:])  # add prompt end

    empty_prompt_ids = np.where(df[columns].isna().all(axis=1).values)[0]

    df["__mdb_prompt"] = ""
    for i in range(len(template)):
        atom = template[i]
        if i < len(columns):
            col = df[columns[i]].replace(to_replace=[None], value="")  # add empty quote if data is missing
            df["__mdb_prompt"] = df["__mdb_prompt"].apply(lambda x: x + atom) + col.astype("string")
        else:
            df["__mdb_prompt"] = df["__mdb_prompt"].apply(lambda x: x + atom)
    prompts = list(df["__mdb_prompt"])

    return prompts, empty_prompt_ids


def get_llm_config(provider: str, args: Dict) -> BaseLLMConfig:
    """
    Helper method that returns the configuration for a given LLM provider.

    :param provider: string with the name of the provider.
    :param config: dictionary with the configuration for the provider.

    :return: LLMConfig object with the configuration for the provider.
    """
    temperature = min(1.0, max(0.0, args.get("temperature", 0.0)))
    if provider == "openai":
        if any(x in args.get("model_name", "") for x in ["o1", "o3"]):
            # for o1 and 03, 'temperature' does not support 0.0 with this model. Only the default (1) value is supported
            temperature = 1

        return OpenAIConfig(
            model_name=args.get("model_name", DEFAULT_OPENAI_MODEL),
            temperature=temperature,
            max_retries=args.get("max_retries", DEFAULT_OPENAI_MAX_RETRIES),
            max_tokens=args.get("max_tokens", DEFAULT_OPENAI_MAX_TOKENS),
            openai_api_base=args.get("base_url", None),
            openai_api_key=args["api_keys"].get("openai", None),
            openai_organization=args.get("api_organization", None),
            request_timeout=args.get("request_timeout", None),
        )
    if provider == "anthropic":
        return AnthropicConfig(
            model=args.get("model_name", DEFAULT_ANTHROPIC_MODEL),
            temperature=temperature,
            max_tokens=args.get("max_tokens", None),
            top_p=args.get("top_p", None),
            top_k=args.get("top_k", None),
            default_request_timeout=args.get("default_request_timeout", None),
            anthropic_api_key=args["api_keys"].get("anthropic", None),
            anthropic_api_url=args.get("base_url", None),
        )
    if provider == "litellm":
        model_kwargs = {
            "api_key": args["api_keys"].get("litellm", None),
            "top_p": args.get("top_p", None),
            "request_timeout": args.get("request_timeout", None),
            "frequency_penalty": args.get("frequency_penalty", None),
            "presence_penalty": args.get("presence_penalty", None),
            "logit_bias": args.get("logit_bias", None),
        }
        return LiteLLMConfig(
            model=args.get("model_name", DEFAULT_LITELLM_MODEL),
            temperature=temperature,
            api_base=args.get("base_url", DEFAULT_LITELLM_BASE_URL),
            max_retries=args.get("max_retries", DEFAULT_OPENAI_MAX_RETRIES),
            max_tokens=args.get("max_tokens", DEFAULT_OPENAI_MAX_TOKENS),
            top_p=args.get("top_p", None),
            top_k=args.get("top_k", None),
            custom_llm_provider=args.get("custom_llm_provider", DEFAULT_LITELLM_PROVIDER),
            model_kwargs=model_kwargs,
        )
    if provider == "ollama":
        return OllamaConfig(
            base_url=args.get("base_url", DEFAULT_OLLAMA_BASE_URL),
            model=args.get("model_name", DEFAULT_OLLAMA_MODEL),
            temperature=temperature,
            top_p=args.get("top_p", None),
            top_k=args.get("top_k", None),
            timeout=args.get("request_timeout", None),
            format=args.get("format", None),
            headers=args.get("headers", None),
            num_predict=args.get("num_predict", None),
            num_ctx=args.get("num_ctx", None),
            num_gpu=args.get("num_gpu", None),
            repeat_penalty=args.get("repeat_penalty", None),
            stop=args.get("stop", None),
            template=args.get("template", None),
        )
    if provider == "nvidia_nim":
        return NvidiaNIMConfig(
            base_url=args.get("base_url", DEFAULT_NVIDIA_NIM_BASE_URL),
            model=args.get("model_name", DEFAULT_NVIDIA_NIM_MODEL),
            temperature=temperature,
            top_p=args.get("top_p", None),
            timeout=args.get("request_timeout", None),
            format=args.get("format", None),
            headers=args.get("headers", None),
            num_predict=args.get("num_predict", None),
            num_ctx=args.get("num_ctx", None),
            num_gpu=args.get("num_gpu", None),
            repeat_penalty=args.get("repeat_penalty", None),
            stop=args.get("stop", None),
            template=args.get("template", None),
            nvidia_api_key=args["api_keys"].get("nvidia_nim", None),
        )
    if provider == "mindsdb":
        return MindsdbConfig(
            model_name=args["model_name"],
            project_name=args.get("project_name", config.get("default_project")),
        )
    if provider == "vllm":
        return OpenAIConfig(
            model_name=args.get("model_name"),
            temperature=temperature,
            max_retries=args.get("max_retries", DEFAULT_OPENAI_MAX_RETRIES),
            max_tokens=args.get("max_tokens", DEFAULT_OPENAI_MAX_TOKENS),
            openai_api_base=args.get("base_url", DEFAULT_VLLM_SERVER_URL),
            openai_api_key=args["api_keys"].get("vllm", "EMPTY`"),
            openai_organization=args.get("api_organization", None),
            request_timeout=args.get("request_timeout", None),
        )
    if provider == "google":
        return GoogleConfig(
            model=args.get("model_name", DEFAULT_GOOGLE_MODEL),
            temperature=temperature,
            top_p=args.get("top_p", None),
            top_k=args.get("top_k", None),
            max_output_tokens=args.get("max_tokens", None),
            google_api_key=args["api_keys"].get("google", None),
        )
    if provider == "writer":
        return WriterConfig(
            model_name=args.get("model_name", "palmyra-x5"),
            temperature=temperature,
            max_tokens=args.get("max_tokens", None),
            top_p=args.get("top_p", None),
            stop=args.get("stop", None),
            best_of=args.get("best_of", None),
            writer_api_key=args["api_keys"].get("writer", None),
            writer_org_id=args.get("writer_org_id", None),
            base_url=args.get("base_url", None),
        )

    raise ValueError(f"Provider {provider} is not supported.")


def ft_jsonl_validation(
    items: list,  # read from a JSONL file
    messages_col: str = "messages",
    # valid keys for each chat message
    role_key: str = "role",
    content_key: str = "content",
    name_key: str = "name",
    # valid roles for each chat message
    system_key: str = "system",
    user_key: str = "user",
    assistant_key: str = "assistant",
):
    """
    This helper checks a list of dictionaries for compliance with the format usually expected by LLM providers
    (such as OpenAI or AnyscaleEndpoints) for fine-tuning LLMs that generate chat completions.

    Defaults are set according to the expected format, but these can be changed if needed by any given provider.

    :param items: list of JSON lines, each dictionary containing a chat sequence. Should be read from a JSONL file.
    :param messages_col: key in each dictionary to access a sequence of chat messages


    For chat-level checks, this method defers to `ft_chat_format_validation()` below. Relevant parameters for it are:

    For each chat:
    :param role_key: key that defines the role of each message (e.g. system, user, or LLM)
    :param content_key: key that defines the content of each message
    :param name_key: key that defines the name of each message

    For each message:
    :param system_key: valid role for each chat message
    :param user_key: valid role for each chat message
    :param assistant_key: valid role for each chat message

    :return: None, raises an Exception if validation fails.
    """  # noqa
    try:
        if not all([isinstance(m, dict) for m in items]):
            raise Exception("Each line in the provided data should be a dictionary")

        for line_num, batch in enumerate(items):
            prefix = f"error in chat #{line_num + 1}, "

            if not isinstance(batch[messages_col], list):
                raise Exception(
                    f"{prefix}Each line in the provided data should have a '{messages_col}' key with a list of messages"
                )  # noqa

            if messages_col not in batch:
                raise Exception(f"{prefix}Each line in the provided data should have a '{messages_col}' key")

            messages = batch[messages_col]
            try:
                ft_chat_format_validation(
                    messages,
                    role_key=role_key,
                    content_key=content_key,
                    name_key=name_key,
                    system_key=system_key,
                    user_key=user_key,
                    assistant_key=assistant_key,
                )
            except Exception as e:
                raise Exception(f"{prefix}{e}")

    except Exception as e:
        raise Exception(f"Fine-tuning data format is not valid. Got {e}")


def ft_chat_format_validation(
    chat: list,
    transitions: Optional[Dict] = None,
    system_key: str = "system",
    user_key: str = "user",
    assistant_key: str = "assistant",
    role_key: str = "role",
    content_key: str = "content",
    name_key: str = "name",
):
    """
    Finite state machine to check a chat has valid format to finetune an LLM with it.
    Follows OpenAI ChatCompletion format (also used by other providers such as AnyscaleEndpoints).
    Reference: https://cookbook.openai.com/examples/chat_finetuning_data_prep

    The unit test in `test_llm_utils.py` for examples of valid and invalid chats.

    :param chat: list of dictionaries, each containing a chat message
    :param transitions: optional dictionary defining valid transitions between chat messages (e.g. from system to user to assistant)

    For each chat:
    :param role_key: key that defines the role of each message (e.g. system, user, or LLM)
    :param content_key: key that defines the content of each message
    :param name_key: key that defines the name of each message

    For each message:
    :param system_key: valid role for each chat message
    :param user_key: valid role for each chat message
    :param assistant_key: valid role for each chat message

    :return: None if chat is valid, otherwise raise an informative Exception.
    """  # noqa

    valid_keys = (role_key, content_key, name_key)
    valid_roles = (system_key, user_key, assistant_key)

    for c in chat:
        if any(k not in valid_keys for k in c.keys()):
            raise Exception(f"Each message should only have these keys: `{valid_keys}`. Found: `{c.keys()}`")

    roles = [m[role_key] for m in chat]
    contents = [m[content_key] for m in chat]

    if len(roles) != len(contents):
        raise Exception(f"Each message should contain both `{role_key}` and `{content_key}` fields")

    if len(roles) == 0:
        raise Exception("Chat should have at least one message")

    if assistant_key not in roles:
        raise Exception("Chat should have at least one assistant message")  # otherwise it is useless for FT

    if user_key not in roles:
        raise Exception("Chat should have at least one user message")  # perhaps remove in the future

    # set default transitions for finite state machine if undefined
    if transitions is None:
        transitions = {
            None: [system_key, user_key],
            system_key: [user_key],
            user_key: [assistant_key],
            assistant_key: [user_key],
        }

    # check order is valid via finite state machine
    state = None
    for i, (role, content) in enumerate(zip(roles, contents)):
        prefix = f"message #{i + 1}: "

        # check invalid roles
        if role not in valid_roles:
            raise Exception(f"{prefix}Invalid role (found `{role}`, expected one of `{valid_roles}`)")

        # check content
        if not isinstance(content, str):
            raise Exception(f"{prefix}Content should be a string, got type `{type(content)}`")

        # check transition
        if role not in transitions[state]:
            raise Exception(f"{prefix}Invalid transition from `{state}` to `{role}`")
        else:
            state = role


def ft_formatter(df: pd.DataFrame) -> List[Dict]:
    """
    Data preparation entry point for chat LLM finetuning. This method will dispatch to the appropriate formatters.

    Supported formats:
        - code: long tabular format with a `code` column
        - chat: long tabular format with `role` and `content` columns, or a JSON format with a `chat_json` column.
    """
    if "code" in df.columns:
        df = ft_code_formatter(df)

    elif {"question", "context", "answer"}.issubset(set(df.columns)):
        # TODO: handler user-specified names for these columns
        df = ft_cqa_formatter(df)

    return ft_chat_formatter(df)


def ft_chat_formatter(df: pd.DataFrame) -> List[Dict]:
    """
    For more details, check `FineTuning -> Data Format` in the Anyscale API reference, or the OpenAI equivalent.
    Additionally, the unit test in `test_llm_utils.py` provides example usage.

    :param df: input dataframe has chats in one of the following formats:
        1) long tabular: at least two columns, `role` and `content`. Rows contain >= 1 chats in long (stacked) format.

        2) JSON: at least one column, `chat_json`. Each row contains exactly 1 chat in JSON format.
            Example for `chat_json` content:
                > `{"messages": [{"role": "user", "content": "Hello!"}, {"role": "assistant", "content": "Hi!"}]}`

    Optional df columns are:
        - chat_id: unique identifier for each chat
        - message_id: unique identifier for each message within each chat

        Data will be sorted by both if they are provided.

        If only `chat_id` is provided, data will be sorted by it with a stable sort, so messages for each chat
        will be in the same order as in the original data.

        If only `message_id` is provided, it must not contain duplicate IDs. Entire dataset will be treated
        as a single chat. Otherwise an exception will be raised.

    :return: list of chats. Each chat is a dictionary with a top level key 'messages' containing a list of messages
    that comply with the OpenAI's ChatEndpoint expected format (i.e., each is a dictionary with a `role` and
    `content` key.

    """  # noqa
    # 1. pre-sort df on optional columns
    if "chat_id" in df.columns:
        if "message_id" in df.columns:
            df = df.sort_values(["chat_id", "message_id"])
        else:
            df = df.sort_values(["chat_id"], kind="stable")
    elif "message_id" in df.columns:
        if df["message_id"].duplicated().any():
            raise Exception("If `message_id` is provided, it must not contain duplicate IDs.")
        df = df.sort_values(["message_id"])

    # 2. build chats
    chats = []

    # 2a. chats are in JSON format
    if "chat_json" in df.columns:
        for _, row in df.iterrows():
            try:
                chat = json.loads(row["chat_json"])
                assert list(chat.keys()) == ["messages"], "Each chat should have a 'messages' key, and nothing else."
                ft_chat_format_validation(chat["messages"])  # will raise Exception if chat is invalid
                chats.append(chat)
            except json.JSONDecodeError:
                pass  # TODO: add logger info here, prompt user to clean dataset carefully

    # 2b. chats are in tabular format - aggregate each chat sequence into one row
    else:
        chat = []
        for i, row in df.iterrows():
            if row["role"] == "system" and len(chat) > 0:
                ft_chat_format_validation(chat)  # will raise Exception if chat is invalid
                chats.append({"messages": chat})
                chat = []
            event = {"role": row["role"], "content": row["content"]}
            chat.append(event)

        ft_chat_format_validation(chat)  # will raise Exception if chat is invalid
        chats.append({"messages": chat})

    return chats


def ft_code_formatter(
    df: pd.DataFrame,
    format="chat",
    language="python",
    chunk_size=100,
    chunk_overlap=0,
    chat_sections=("Code prefix", "Code suffix", "Completion"),
    fim_tokens=("<PRE>", "<SUF>", "<MID>"),
) -> pd.DataFrame:
    """
    This utility processes a raw codebase stored as a dataframe with a `code` column, where
    every row may be an entire file or some portion of it.
    It chunks code into triples made of a prefix, middle, and suffix.

    Depending on the target LLM, these triples are then formatted into a chat-like prompt, or a
    fill-in-the-middle (FIM) prompt. The latter is used for fine-tuning models like codellama,
    while the former is more generic and should work with any LLM that supports the ChatCompletion
    format, as the rest of our tools do.
    """

    # input and setup validation
    assert len(df) > 0, "Input dataframe should not be empty"
    assert "code" in df.columns, "Input dataframe should have a 'code' column"
    assert chunk_size > 0 and isinstance(chunk_size, int), "`chunk_size` should be a positive integer"

    supported_formats = ["chat", "fim"]
    supported_langs = [e.value for e in Language]
    assert language.lower() in supported_langs, f"Invalid language. Valid choices are: {supported_langs}"

    # ensure correct encoding
    df["code"] = df["code"].map(lambda x: x.encode("utf8").decode("unicode_escape"))

    # set prompt templates
    system_prompt = "You are a powerful text to code model. Your job is to provide great code completions. As context, you are given code that is found immediately before and after the code you must generate.\n\nYou must output the code that should go in between the prefix and suffix.\n\n"
    if format == "chat":
        templates = [f"### {c}:" for c in chat_sections]
    elif format == "fim":
        templates = fim_tokens
    else:
        raise Exception(f"Invalid format. Please choose one of {supported_formats}")

    # split code into chunks
    code_splitter = RecursiveCharacterTextSplitter.from_language(
        language=getattr(Language, language.upper()),
        chunk_size=3 * chunk_size,  # each triplet element has `chunk_size`
        chunk_overlap=chunk_overlap,  # some overlap here is fine
    )
    chunk_docs = code_splitter.create_documents(list(df["code"]))
    chunks = [c.page_content for c in chunk_docs]

    # split each chunk into a triplet, with no overlap
    triplet_splitter = RecursiveCharacterTextSplitter.from_language(
        language=getattr(Language, language.upper()),
        chunk_size=chunk_size,
        chunk_overlap=0,  # no overlap admitted, otherwise context may leak into answer
    )
    triplet_chunk_docs = triplet_splitter.create_documents(chunks)
    chunks = [c.page_content for c in triplet_chunk_docs]
    chunks = chunks[: len(chunks) - len(chunks) % 3]  # should be a multiple of 3

    # format chunks into prompts
    roles = []
    contents = []
    for idx in range(0, len(chunks), 3):
        pre, mid, suf = chunks[idx : idx + 3]
        interleaved = list(itertools.chain(*zip(templates, (pre, suf, mid))))
        user = "\n".join(interleaved[:-1])
        assistant = "\n".join(interleaved[-1:])
        roles.extend(["system", "user", "assistant"])
        contents.extend([system_prompt, user, assistant])

    # return formatted prompts in a dataframe to be processed by `ft_chat_formatter()`
    df = pd.DataFrame({"role": roles, "content": contents})
    return df


def ft_cqa_formatter(
    df: pd.DataFrame,
    question_col="question",
    answer_col="answer",
    instruction_col="instruction",
    context_col="context",
    default_instruction="You are a helpful assistant.",
    default_context="",
) -> pd.DataFrame:
    # input and setup validation
    assert len(df) > 0, "Input dataframe should not be empty"
    assert {question_col, answer_col}.issubset(set(df.columns)), (
        f"Input dataframe must have columns `{question_col}`, and `{answer_col}`"
    )  # noqa

    if instruction_col not in df.columns:
        df[instruction_col] = default_instruction

    if context_col not in df.columns:
        df[context_col] = default_context

    # format data into chat-like prompts
    roles = []
    contents = []
    for i, row in df.iterrows():
        system = "\n".join([row[instruction_col], row[context_col]])
        user = row[question_col]
        assistant = row[answer_col]
        roles.extend(["system", "user", "assistant"])
        contents.extend([system, user, assistant])

    return pd.DataFrame({"role": roles, "content": contents})



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Apache Derby server/database.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect to Apache Derby.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': """
            The database name to use when connecting with the Apache Derby server.
        """
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Apache Derby server. If specified this is also treated as the schema.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Apache Derby server.',
        'secret': True
    },
    jdbcClass={
        'type': ARG_TYPE.STR,
        'description': 'The jdbc class which should be used to establish the connection, the default value is:  org.apache.derby.jdbc.ClientDriver.'
    },
    jdbcJarLocation={
        'type': ARG_TYPE.STR,
        'description': 'The location of the jar files which contain the JDBC class. This need not be specified if the required classes are already added to the CLASSPATH variable.'
    }

)


connection_args_example = OrderedDict(
    host='localhost',
    port='1527',
    user='test',
    password='test',
    database="testdb",
    jdbcClass='org.apache.derby.jdbc.ClientDriver',
    jdbcJarLocation='/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derby.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyclient.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbynet.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyoptionaltools.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyrun.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyshared.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbytools.jar',
)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import datetime as dt
import struct

import pandas as pd

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import TYPES


class BinaryResultsetRowPacket(Packet):
    """
    Implementation based on:
    https://mariadb.com/kb/en/resultset-row/#binary-resultset-row
    https://dev.mysql.com/doc/internals/en/null-bitmap.html
    """

    def setup(self):
        data = self._kwargs.get("data", {})
        columns = self._kwargs.get("columns", {})

        self.value = [b"\x00"]

        # NOTE: according to mysql's doc offset=0 only for COM_STMT_EXECUTE, mariadb's doc does't mention that
        # but in fact it looks like offset=2 everywhere
        offset = 2
        nulls_bitmap = bytearray((len(columns) + offset + 7) // 8)
        for i, el in enumerate(data):
            if el is not None:
                continue
            byte_index = (i + offset) // 8
            bit_index = (i + offset) % 8
            nulls_bitmap[byte_index] |= 1 << bit_index
        self.value.append(bytes(nulls_bitmap))

        for i, col in enumerate(columns):
            # NOTE at this moment all types sends as strings, and it works
            val = data[i]
            if val is None:
                continue

            enc = None
            env_val = None
            col_type = col["type"]
            if col_type == TYPES.MYSQL_TYPE_DOUBLE:
                enc = "<d"
                val = float(val)
            elif col_type == TYPES.MYSQL_TYPE_LONGLONG:
                enc = "<q"
                val = int(val)
            elif col_type == TYPES.MYSQL_TYPE_LONG:
                enc = "<l"
                val = int(val)
            elif col_type == TYPES.MYSQL_TYPE_FLOAT:
                enc = "<f"
                val = float(val)
            elif col_type == TYPES.MYSQL_TYPE_YEAR:
                enc = "<h"
                val = int(float(val))
            elif col_type == TYPES.MYSQL_TYPE_SHORT:
                enc = "<h"
                val = int(val)
            elif col_type == TYPES.MYSQL_TYPE_TINY:
                enc = "<B"
                val = int(val)
            elif col_type == TYPES.MYSQL_TYPE_DATE:
                env_val = self.encode_date(val)
            elif col_type == TYPES.MYSQL_TYPE_TIMESTAMP:
                env_val = self.encode_date(val)
            elif col_type == TYPES.MYSQL_TYPE_DATETIME:
                env_val = self.encode_date(val)
            elif col_type == TYPES.MYSQL_TYPE_TIME:
                env_val = self.encode_time(val)
            elif col_type == TYPES.MYSQL_TYPE_NEWDECIMAL:
                enc = "string"
            elif col_type == TYPES.MYSQL_TYPE_VECTOR:
                enc = "byte"
            elif col_type == TYPES.MYSQL_TYPE_JSON:
                # json have to be encoded as byte<lenenc>, but actually for json there is no differ with string<>
                enc = "string"
            else:
                enc = "string"

            if enc == "":
                raise Exception(f"Column with type {col_type} cant be encripted")

            if enc == "byte":
                self.value.append(Datum("string", val, "lenenc").toStringPacket())
            elif enc == "string":
                if not isinstance(val, str):
                    val = str(val)
                self.value.append(Datum("string", val, "lenenc").toStringPacket())
            else:
                if env_val is None:
                    env_val = struct.pack(enc, val)
                self.value.append(env_val)

    def encode_time(self, val: dt.time | str) -> bytes:
        """https://mariadb.com/kb/en/resultset-row/#time-binary-encoding"""
        if isinstance(val, str):
            try:
                val = dt.datetime.strptime(val, "%H:%M:%S").time()
            except ValueError:
                val = dt.datetime.strptime(val, "%H:%M:%S.%f").time()
        if val == dt.time(0, 0, 0):
            return struct.pack("<B", 0)  # special case for 0 time
        out = struct.pack("<B", 0)  # positive time
        out += struct.pack("<L", 0)  # days
        out += struct.pack("<B", val.hour)
        out += struct.pack("<B", val.minute)
        out += struct.pack("<B", val.second)
        if val.microsecond > 0:
            out += struct.pack("<L", val.microsecond)
            len_bit = struct.pack("<B", 12)
        else:
            len_bit = struct.pack("<B", 8)
        return len_bit + out

    def encode_date(self, val):
        # date_type = None
        # date_value = None

        if isinstance(val, str):
            forms = [
                "%Y-%m-%d",
                "%Y-%m-%d %H:%M:%S",
                "%Y-%m-%d %H:%M:%S.%f",
                "%Y-%m-%dT%H:%M:%S",
                "%Y-%m-%dT%H:%M:%S.%f",
            ]
            for f in forms:
                try:
                    date_value = dt.datetime.strptime(val, f)
                    break
                except ValueError:
                    date_value = None
            if date_value is None:
                raise ValueError(f"Invalid date format: {val}")
            date_type = "datetime"
        elif isinstance(val, pd.Timestamp):
            date_value = val
            date_type = "datetime"

        out = struct.pack("<H", date_value.year)
        out += struct.pack("<B", date_value.month)
        out += struct.pack("<B", date_value.day)

        if date_type == "datetime":
            out += struct.pack("<B", date_value.hour)
            out += struct.pack("<B", date_value.minute)
            out += struct.pack("<B", date_value.second)
            out += struct.pack("<L", date_value.microsecond)

        len_bit = struct.pack("<B", len(out))
        return len_bit + out

    @property
    def body(self):
        string = b"".join(self.value)
        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint

        pprint.pprint(str(BinaryResultsetRowPacket().get_packet_string()))


if __name__ == "__main__":
    BinaryResultsetRowPacket.test()



"""updated data catalog data types

Revision ID: 608e376c19a7
Revises: a44643042fe8
Create Date: 2025-06-09 23:20:34.739735

"""

from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = "608e376c19a7"
down_revision = "a44643042fe8"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("meta_tables", schema=None) as batch_op:
        batch_op.alter_column(
            "row_count",
            type_=sa.BigInteger(),
            existing_type=sa.Integer(),
            existing_nullable=True,
            existing_server_default=None,
        )

    with op.batch_alter_table("meta_column_statistics", schema=None) as batch_op:
        batch_op.alter_column(
            "distinct_values_count",
            type_=sa.BigInteger(),
            existing_type=sa.Integer(),
            existing_nullable=True,
            existing_server_default=None,
        )


def downgrade():
    with op.batch_alter_table("meta_tables", schema=None) as batch_op:
        batch_op.alter_column(
            "row_count",
            type_=sa.Integer(),
            existing_type=sa.BigInteger(),
            existing_nullable=True,
            existing_server_default=None,
        )

    with op.batch_alter_table("meta_column_statistics", schema=None) as batch_op:
        batch_op.alter_column(
            "distinct_values_count",
            type_=sa.Integer(),
            existing_type=sa.BigInteger(),
            existing_nullable=True,
            existing_server_default=None,
        )



import json
import math
import re
import time
from typing import List

import pandas as pd
import datetime as dt

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb_sql_parser import Identifier, Select, Constant, Star, parse_sql, BinaryOperation
from mindsdb.utilities import log

from mindsdb.interfaces.knowledge_base.llm_client import LLMClient

logger = log.getLogger(__name__)


GENERATE_QA_SYSTEM_PROMPT = """
Your task is to generate question and answer pairs for a search engine.
The search engine will take your query and return a list of documents.
You will be given a text and you need to generate a question that can be answered using the information in the text.
Your questions will be used to evaluate the search engine.
Question should always have enough clues to identify the specific text that this question is generated from.
Never ask questions like "What license number is associated with Amend 6" because Amend 6 could be found in many documents and the question is not specific enough.
Example output 1:  {\"query\": \"What processor does the HP 2023 14\" FHD IPS Laptop use?\", \"reference_answer\": \"Ryzen 3 5300U\"}
Example output 2: {\"query\": \"What is the name of the river in Paris?\", \"reference_answer\": \"Seine\"}
Don't generate questions like "What is being amended in the application?" because these questions cannot be answered using the text and without knowing which document it refers to.
The question should be answerable without the text, but the answer should be present in the text.
Return ONLY a json response. No other text.
"""


def calc_entropy(values: List[float]) -> float:
    """
    Alternative of scipy.stats.entropy, to not add `scipy` dependency
    :param values: Input distribution
    :return: The calculated entropy.
    """
    # normalize & filter
    total = sum(values)
    values = [i / total for i in values if i > 0]
    # calc
    return -sum([pk * math.log(pk) for pk in values])


def sanitize_json_response(response: str) -> str:
    """Remove markdown code block formatting from JSON response and extract valid JSON."""
    if not response or not response.strip():
        raise ValueError("Empty response provided.")

    # Remove leading/trailing whitespace
    response = response.strip()

    # Remove markdown code block markers if present
    response = re.sub(r"^```(?:json|JSON)?\s*", "", response, flags=re.MULTILINE)
    response = re.sub(r"\s*```$", "", response, flags=re.MULTILINE)
    response = response.strip()

    # Find the first opening brace
    start_idx = response.find("{")
    if start_idx == -1:
        raise ValueError("No JSON object found in the response.")

    # Try to parse JSON starting from first { with increasing end positions
    # This handles nested objects and strings with braces correctly
    for end_idx in range(len(response), start_idx, -1):  # Start from end and work backwards
        candidate = response[start_idx:end_idx]
        try:
            parsed = json.loads(candidate)
            # Ensure it's a dictionary (object) not just any valid JSON
            if isinstance(parsed, dict):
                return candidate
        except json.JSONDecodeError:
            continue

    raise ValueError("No valid JSON object found in the response.")


class EvaluateBase:
    DEFAULT_QUESTION_COUNT = 20
    DEFAULT_SAMPLE_SIZE = 10000

    def __init__(self, session, knowledge_base):
        self.kb = knowledge_base
        self.name = knowledge_base._kb.name
        self.session = session

        self._llm_client = None

    def generate(self, sampled_df: pd.DataFrame) -> pd.DataFrame:
        # generate test data from sample
        raise NotImplementedError

    def evaluate(self, test_data: pd.DataFrame) -> pd.DataFrame:
        # create evaluate metric from test data
        raise NotImplementedError

    def _set_llm_client(self, llm_params: dict):
        """
        Logic to get LLM setting:
        - first get `llm` setting of ‘evaluate’ command
        - if not defined, look at the knowledge base reranker config
        """
        if llm_params is None:
            llm_params = self.kb._kb.params.get("reranking_model")

        self.llm_client = LLMClient(llm_params)

    def generate_test_data(self, gen_params: dict) -> pd.DataFrame:
        # Extract source data (from users query or from KB itself) and call `generate` to get test data

        if "from_sql" in gen_params:
            # get data from sql
            query = parse_sql(gen_params["from_sql"])
            if not isinstance(query, Select) or query.from_table is None:
                raise ValueError(f"Query not supported {gen_params['from_sql']}")

            dn, table_name = self._get_dn_table(query.from_table)
            query.from_table = table_name
            if query.limit is None:
                query.limit = Constant(self.DEFAULT_SAMPLE_SIZE)

            response = dn.query(query=query, session=self.session)
            df = response.data_frame

            if "content" not in df.columns:
                raise ValueError(f"`content` column isn't found in provided sql: {gen_params['from_sql']}")

            df.rename(columns={"content": "chunk_content"}, inplace=True)
        else:
            # get data from knowledge base
            df = self.kb.select_query(
                Select(
                    targets=[Identifier("chunk_content"), Identifier("id")], limit=Constant(self.DEFAULT_SAMPLE_SIZE)
                )
            )

        if "count" in gen_params:
            number_of_questions = gen_params["count"]
        else:
            number_of_questions = self.DEFAULT_QUESTION_COUNT

        number_of_questions = min(number_of_questions, len(df))
        sampled_df = df.sample(n=number_of_questions)

        return self.generate(sampled_df)

    def read_from_table(self, test_table: Identifier) -> pd.DataFrame:
        # read data from table

        dn, table_name = self._get_dn_table(test_table)

        query = Select(
            targets=[Star()],
            from_table=table_name,
        )
        response = dn.query(query=query, session=self.session)
        return response.data_frame

    def _get_dn_table(self, table_name: Identifier):
        if len(table_name.parts) < 2:
            raise ValueError(f"Can't find database, table name must have at least 2 parts: {table_name}")

        integration_name = table_name.parts[0]
        table_name = Identifier(parts=table_name.parts[1:])
        dn = self.session.datahub.get(integration_name)
        if dn is None:
            raise ValueError(f"Can't find database: {integration_name}")
        return dn, table_name

    def save_to_table(self, table_name: Identifier, df: pd.DataFrame, is_replace=False):
        # save data to table

        dn, table_name = self._get_dn_table(table_name)

        data = ResultSet.from_df(df)

        dn.create_table(
            table_name=table_name,
            result_set=data,
            is_replace=is_replace,
            is_create=True,
            raise_if_exists=False,
        )

    def run_evaluate(self, params: dict) -> pd.DataFrame:
        # evaluate function entry point

        self._set_llm_client(params.get("llm"))

        if "test_table" not in params:
            raise ValueError('The table with  has to be defined in "test_table" parameter')

        test_table = params["test_table"]

        if isinstance(test_table, str):
            test_table = Identifier(test_table)

        if "generate_data" in params:
            # generate question / answers using llm
            gen_params = params["generate_data"]
            if not isinstance(gen_params, dict):
                gen_params = {}
            test_data = self.generate_test_data(gen_params)

            self.save_to_table(test_table, test_data, is_replace=True)

        if params.get("evaluate", True) is False:
            # no evaluate is required
            return pd.DataFrame()

        test_data = self.read_from_table(test_table)

        scores = self.evaluate(test_data)
        scores["id"] = math.floor(time.time())  # unique ID for the evaluation run
        scores["name"] = self.name
        scores["created_at"] = dt.datetime.now()

        # save scores
        if "save_to" in params:
            to_table = params["save_to"]
            if isinstance(to_table, str):
                to_table = Identifier(to_table)
            self.save_to_table(to_table, scores.copy())

        return scores

    @staticmethod
    def run(session, kb_table, params) -> pd.DataFrame:
        # choose the evaluator version according to the 'version' parameter in config

        evaluate_version = params.get("version", "doc_id")

        if evaluate_version == "llm_relevancy":
            cls = EvaluateRerank
        elif evaluate_version == "doc_id":
            cls = EvaluateDocID
        else:
            raise NotImplementedError(f"Version of evaluator is not implemented: {evaluate_version}")

        return cls(session, kb_table).run_evaluate(params)


class EvaluateRerank(EvaluateBase):
    """
    Rank responses from KB using LLM (by calling KB reranker function)
    """

    TOP_K = 10

    def generate(self, sampled_df: pd.DataFrame) -> pd.DataFrame:
        qa_data = []
        count_errors = 0
        for chunk_content in sampled_df["chunk_content"]:
            try:
                question, answer = self.generate_question_answer(chunk_content)
            except ValueError as e:
                # allow some numbers of error
                count_errors += 1
                if count_errors > 5:
                    raise e
                continue

            qa_data.append({"text": chunk_content, "question": question, "answer": answer})

        df = pd.DataFrame(qa_data)
        df["id"] = df.index
        return df

    def generate_question_answer(self, text: str) -> (str, str):
        messages = [
            {"role": "system", "content": GENERATE_QA_SYSTEM_PROMPT},
            {"role": "user", "content": f"\n\nText:\n{text}\n\n"},
        ]
        answer = self.llm_client.completion(messages, json_output=True)

        # Sanitize the response by removing markdown code block formatting like ```json
        sanitized_answer = sanitize_json_response(answer)

        try:
            output = json.loads(sanitized_answer)
        except json.JSONDecodeError:
            raise ValueError(f"Could not parse response from LLM: {answer}")

        if "query" not in output or "reference_answer" not in output:
            raise ValueError("Cant find question/answer in LLM response")

        return output.get("query"), output.get("reference_answer")

    def evaluate(self, test_data: pd.DataFrame) -> pd.DataFrame:
        json_to_log_list = []
        questions = test_data.to_dict("records")

        for i, item in enumerate(questions):
            question = item["question"]
            ground_truth = item["answer"]

            start_time = time.time()
            logger.debug(f"Querying [{i + 1}/{len(questions)}]: {question}")
            df_answers = self.kb.select_query(
                Select(
                    targets=[Identifier("chunk_content")],
                    where=BinaryOperation(op="=", args=[Identifier("content"), Constant(question)]),
                    limit=Constant(self.TOP_K),
                )
            )
            query_time = time.time() - start_time

            proposed_responses = list(df_answers["chunk_content"])

            # generate answer using llm
            relevance_score_list = self.kb.score_documents(question, proposed_responses, self.llm_client.params)

            # set binary relevancy
            binary_relevancy_list = [1 if score >= 0.5 else 0 for score in relevance_score_list]

            # calculate first relevant position
            first_relevant_position = next((i for i, x in enumerate(binary_relevancy_list) if x == 1), None)
            json_to_log = {
                "question": question,
                "ground_truth": ground_truth,
                # "relevancy_at_k": relevancy_at_k,
                "binary_relevancy_list": binary_relevancy_list,
                "relevance_score_list": relevance_score_list,
                "first_relevant_position": first_relevant_position,
                "query_time": query_time,
            }
            json_to_log_list.append(json_to_log)

        evaluation_results = self.evaluate_retrieval_metrics(json_to_log_list)
        return pd.DataFrame([evaluation_results])

    def evaluate_retrieval_metrics(self, json_to_log_list):
        """
        Computes retrieval evaluation metrics from the result log.

        Metrics computed:
            - Average Relevancy (mean soft relevance score)
            - Average Relevancy@k (soft score)
            - Average First Relevant Position
            - Mean Reciprocal Rank (MRR)
            - Hit@k
            - Binary Precision@k
            - Average Entropy of Relevance Scores
            - Average nDCG

        Args:
            json_to_log_list (list): List of evaluation logs per query.

        Returns:
            dict: A dictionary containing all computed metrics.
        """

        mrr_list = []
        hit_at_k_matrix = []
        binary_precision_at_k_matrix = []
        ndcg_list = []
        entropy_list = []

        total_relevancy = 0
        relevance_score_matrix = []
        first_relevant_positions = []

        for item in json_to_log_list:
            binary_relevancy = item["binary_relevancy_list"]
            relevance_scores = item["relevance_score_list"]

            # Skip if empty
            if not relevance_scores:
                continue

            # Mean relevancy per query
            query_relevancy = sum(relevance_scores) / len(relevance_scores)
            total_relevancy += query_relevancy

            # Build score matrix for later average@k
            relevance_score_matrix.append(relevance_scores)

            # First relevant position
            pos = item["first_relevant_position"]
            if pos is not None:
                first_relevant_positions.append(pos)

            # MRR
            reciprocal_rank = 1 / (pos + 1) if pos is not None else 0
            mrr_list.append(reciprocal_rank)

            # Hit@k and Binary Precision@k
            hit_row = []
            precision_row = []
            for k in range(1, len(binary_relevancy) + 1):
                hit = int(any(binary_relevancy[:k]))
                precision = sum(binary_relevancy[:k]) / k
                hit_row.append(hit)
                precision_row.append(precision)
            hit_at_k_matrix.append(hit_row)
            binary_precision_at_k_matrix.append(precision_row)

            # Entropy

            entropy = calc_entropy(relevance_scores) if len(relevance_scores) > 1 else 0
            entropy_list.append(entropy)

            # nDCG
            def dcg(scores):
                return sum(score / math.log2(idx + 2) for idx, score in enumerate(scores))

            ideal = sorted(relevance_scores, reverse=True)
            actual_dcg = dcg(relevance_scores)
            ideal_dcg = dcg(ideal)
            ndcg = actual_dcg / ideal_dcg if ideal_dcg > 0 else 0
            ndcg_list.append(ndcg)

        # Aggregated metrics
        num_queries = len(json_to_log_list)
        average_relevancy = total_relevancy / num_queries if num_queries else 0

        # Relevancy@k
        average_relevance_score_by_k = []
        if relevance_score_matrix:
            relevance_score_matrix = list(zip(*relevance_score_matrix))
            for col in relevance_score_matrix:
                avg_k = sum(col) / len(col)
                average_relevance_score_by_k.append(round(avg_k, 2))

        average_first_relevant_position = (
            sum(first_relevant_positions) / len(first_relevant_positions) if first_relevant_positions else None
        )

        mean_mrr = sum(mrr_list) / len(mrr_list) if mrr_list else 0
        hit_at_k_avg = [round(sum(col) / len(col), 2) for col in zip(*hit_at_k_matrix)] if hit_at_k_matrix else []
        binary_precision_at_k_avg = (
            [round(sum(col) / len(col), 2) for col in zip(*binary_precision_at_k_matrix)]
            if binary_precision_at_k_matrix
            else []
        )
        avg_entropy = sum(entropy_list) / len(entropy_list) if entropy_list else 0
        avg_ndcg = sum(ndcg_list) / len(ndcg_list) if ndcg_list else 0

        avg_query_time = sum(item["query_time"] for item in json_to_log_list) / num_queries

        return {
            "avg_relevancy": average_relevancy,
            "avg_relevance_score_by_k": average_relevance_score_by_k,
            "avg_first_relevant_position": average_first_relevant_position,
            "mean_mrr": mean_mrr,
            "hit_at_k": hit_at_k_avg,
            "bin_precision_at_k": binary_precision_at_k_avg,
            "avg_entropy": avg_entropy,
            "avg_ndcg": avg_ndcg,
            "avg_query_time": avg_query_time,
        }


class EvaluateDocID(EvaluateBase):
    """
    Checks if ID in response from KB is matched with doc ID in test dataset
    """

    TOP_K = 20

    def generate(self, sampled_df: pd.DataFrame) -> pd.DataFrame:
        if "id" not in sampled_df.columns:
            raise ValueError("'id' column is required for generating test dataset")

        qa_data = []
        count_errors = 0
        for _, item in sampled_df.iterrows():
            chunk_content = item["chunk_content"]
            try:
                question, answer = self.generate_question_answer(chunk_content)
            except ValueError as e:
                # allow some numbers of error
                count_errors += 1
                if count_errors > 5:
                    raise e
                continue

            qa_data.append({"text": chunk_content, "question": question, "answer": answer, "doc_id": item["id"]})
        if len(qa_data) == 0:
            raise ValueError("No data in generated test dataset")
        df = pd.DataFrame(qa_data)
        return df

    def generate_question_answer(self, text: str) -> (str, str):
        messages = [
            {"role": "system", "content": GENERATE_QA_SYSTEM_PROMPT},
            {"role": "user", "content": f"\n\nText:\n{text}\n\n"},
        ]
        answer = self.llm_client.completion(messages, json_output=True)

        # Sanitize the response by removing markdown code block formatting like ```json
        sanitized_answer = sanitize_json_response(answer)

        try:
            output = json.loads(sanitized_answer)
        except json.JSONDecodeError:
            raise ValueError(f"Could not parse response from LLM: {answer}")

        if "query" not in output or "reference_answer" not in output:
            raise ValueError("Cant find question/answer in LLM response")

        return output.get("query"), output.get("reference_answer")

    def evaluate(self, test_data: pd.DataFrame) -> pd.DataFrame:
        stats = []
        questions = test_data.to_dict("records")

        for i, item in enumerate(questions):
            question = item["question"]
            doc_id = item["doc_id"]

            start_time = time.time()
            logger.debug(f"Querying [{i + 1}/{len(questions)}]: {question}")
            df_answers = self.kb.select_query(
                Select(
                    targets=[Identifier("chunk_content"), Identifier("id")],
                    where=BinaryOperation(op="=", args=[Identifier("content"), Constant(question)]),
                    limit=Constant(self.TOP_K),
                )
            )
            query_time = time.time() - start_time

            retrieved_doc_ids = list(df_answers["id"])

            if doc_id in retrieved_doc_ids:
                doc_found = True
                doc_position = retrieved_doc_ids.index(doc_id)
            else:
                doc_found = False
                doc_position = -1

            stats.append(
                {
                    "question": question,
                    "doc_id": doc_id,
                    "doc_found": doc_found,
                    "doc_position": doc_position,
                    "query_time": query_time,
                }
            )

        evaluation_results = self.summarize_results(stats)
        return pd.DataFrame([evaluation_results])

    def summarize_results(self, stats):
        total_questions = len(stats)
        total_found = sum([1 for stat in stats if stat["doc_found"]])

        accurate_in_top_10 = sum([1 for stat in stats if stat["doc_found"] and stat["doc_position"] < 10])

        # calculate recall curve by position
        recall_curve = {}
        for i in range(self.TOP_K):
            recall_curve[i] = sum([1 for stat in stats if stat["doc_found"] and stat["doc_position"] == i])
        # convert to proportion of total questions
        for i in range(self.TOP_K):
            recall_curve[i] = recall_curve[i] / total_questions
        # calculate cumulative recall
        cumulative_recall = {}
        for i in range(self.TOP_K):
            cumulative_recall[i] = sum([recall_curve[j] for j in range(i + 1)])

        avg_query_time = sum(item["query_time"] for item in stats) / total_questions
        return {
            "total": total_questions,
            "total_found": total_found,
            "retrieved_in_top_10": accurate_in_top_10,
            "cumulative_recall": json.dumps(cumulative_recall),
            "avg_query_time": avg_query_time,
        }



import os
from qbosdk import QuickbooksOnlineSDK
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse
)
from mindsdb.utilities.config import Config
from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql
from .quickbooks_table import AccountsTable, PurchasesTable, BillPaymentsTable, VendorsTable, BillsTable, EmployeesTable

logger = log.getLogger(__name__)


class QuickbooksHandler(APIHandler):
    """
        A class for handling connections and interactions with Quickbooks API.
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get('connection_data', {})

        self.connection_args = {}
        handler_config = Config().get('quickbooks_handler', {})
        for k in ['client_id', 'client_secret', 'refresh_token', 'realm_id', 'environment']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'QUICKBOOKS_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'QUICKBOOKS_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.quickbooks = None
        self.is_connected = False

        accountso = AccountsTable(self)
        self._register_table('accountso', accountso)
        purchases = PurchasesTable(self)
        self._register_table('purchases', purchases)
        bills_payments = BillPaymentsTable(self)
        self._register_table('bills_payments', bills_payments)
        vendors = VendorsTable(self)
        self._register_table('vendors', vendors)
        bills = BillsTable(self)
        self._register_table('bills', bills)
        employees = EmployeesTable(self)
        self._register_table('employees', employees)

    def connect(self):
        if self.is_connected is True:
            return self.quickbooks

        self.quickbooks = QuickbooksOnlineSDK(
            client_id=self.connection_args['client_id'],
            client_secret=self.connection_args['client_secret'],
            realm_id=self.connection_args['realm_id'],
            refresh_token=self.connection_args['refresh_token'],
            environment=self.connection_args['environment']
        )
        self.is_connected = True
        return self.quickbooks

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)

        try:
            quickbooks = self.connect()
            quickbooks.accounts.get()
            logger.info(quickbooks.accounts.get())
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Quickbooks API: {e}. '
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



import gitlab

from mindsdb.integrations.handlers.gitlab_handler.gitlab_tables import GitlabIssuesTable, GitlabMergeRequestsTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

logger = log.getLogger(__name__)


class GitlabHandler(APIHandler):
    """The GitLab handler implementation"""

    def __init__(self, name: str, **kwargs):
        """ constructor
        Args:
            name (str): the handler name
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.repository = connection_data["repository"]
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        gitlab_issues_data = GitlabIssuesTable(self)
        gitlab_merge_requests_data = GitlabMergeRequestsTable(self)
        self._register_table("issues", gitlab_issues_data)
        self._register_table("merge_requests", gitlab_merge_requests_data)

    def connect(self) -> StatusResponse:
        """ Set up the connections required by the handler
        Returns:
            HandlerStatusResponse
        """

        connection_kwargs = {}

        if self.connection_data.get("api_key", None):
            connection_kwargs["private_token"] = self.connection_data["api_key"]

        self.connection = gitlab.Gitlab(**connection_kwargs)
        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)

        try:
            self.connect()
            if self.connection_data.get("api_key", None):
                logger.info("Authenticated as user")
            else:
                logger.info("Proceeding without an API key")

            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to GitLab API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process raw query.
        Args:
            query (str): query in a native format
        Returns:
            HandlerResponse
        """
        ast = parse_sql(query)
        return self.query(ast)



from textwrap import indent


class BaseEntityException(Exception):
    """Base exception for entitys errors

    Attributes:
        message (str): error message
        entity_name (str): entity name
    """

    def __init__(self, message: str, entity_name: str = None) -> None:
        self.message = message
        self.entity_name = entity_name or "unknown"

    def __str__(self) -> str:
        return f"{self.message}: {self.entity_name}"


class EntityExistsError(BaseEntityException):
    """Raise when entity exists, but should not"""

    def __init__(self, message: str = None, entity_name: str = None) -> None:
        if message is None:
            message = "Entity exists error"
        super().__init__(message, entity_name)


class EntityNotExistsError(BaseEntityException):
    """Raise when entity not exists, but should"""

    def __init__(self, message: str = None, entity_name: str = None) -> None:
        if message is None:
            message = "Entity does not exists error"
        super().__init__(message, entity_name)


def format_db_error_message(
    db_name: str | None = None,
    db_type: str | None = None,
    db_error_msg: str | None = None,
    failed_query: str | None = None,
    is_external: bool = True,
) -> str:
    """Format the error message for the database query.

    Args:
        db_name (str | None): The name of the database.
        db_type (str | None): The type of the database.
        db_error_msg (str | None): The error message.
        failed_query (str | None): The failed query.
        is_external (bool): True if error appeared in external database, False if in internal duckdb

    Returns:
        str: The formatted error message.
    """
    error_message = "Failed to execute external database query during query processing."
    if is_external:
        error_message = (
            "An error occurred while executing a derived query on the external "
            "database during processing of your original SQL query."
        )
    else:
        error_message = (
            "An error occurred while processing an internally generated query derived from your original SQL statement."
        )
    if db_name is not None or db_type is not None:
        error_message += "\n\nDatabase Details:"
        if db_name is not None:
            error_message += f"\n- Name: {db_name}"
        if db_type is not None:
            error_message += f"\n- Type: {db_type}"

    if db_error_msg is not None:
        error_message += f"\n\nError:\n{indent(db_error_msg, '    ')}"

    if failed_query is not None:
        error_message += f"\n\nFailed Query:\n{indent(failed_query, '    ')}"

    return error_message



from concurrent.futures import as_completed, TimeoutError
from typing import Optional, Dict
import re

from langchain.agents import AgentExecutor
from langchain.agents.initialize import initialize_agent
from langchain.chains.conversation.memory import ConversationSummaryBufferMemory
from langchain_core.prompts import PromptTemplate
from langchain_core.messages import SystemMessage, HumanMessage

import numpy as np
import pandas as pd

from mindsdb.interfaces.agents.safe_output_parser import SafeOutputParser
from mindsdb.interfaces.agents.langchain_agent import (
    get_llm_provider,
    get_embedding_model_provider,
    create_chat_model,
    get_chat_model_params,
)

from mindsdb.interfaces.agents.constants import (
    DEFAULT_AGENT_TIMEOUT_SECONDS,
    DEFAULT_AGENT_TOOLS,
    DEFAULT_AGENT_TYPE,
    DEFAULT_MAX_ITERATIONS,
    DEFAULT_MAX_TOKENS,
    DEFAULT_MODEL_NAME,
    USER_COLUMN,
    ASSISTANT_COLUMN,
)
from mindsdb.integrations.utilities.rag.settings import DEFAULT_RAG_PROMPT_TEMPLATE
from mindsdb.integrations.handlers.langchain_handler.tools import setup_tools
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.interfaces.storage.model_fs import HandlerStorage, ModelStorage
from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import (
    construct_model_from_args,
)
from mindsdb.integrations.handlers.openai_handler.constants import CHAT_MODELS_PREFIXES  # noqa: F401 - for dependency checker

from mindsdb.utilities import log
from mindsdb.utilities.context_executor import ContextThreadPoolExecutor

_PARSING_ERROR_PREFIXES = ["An output parsing error occured", "Could not parse LLM output"]

logger = log.getLogger(__name__)


class LangChainHandler(BaseMLEngine):
    """
    This is a MindsDB integration for the LangChain library, which provides a unified interface for interacting with
    various large language models (LLMs).

    Supported LLM providers:
        - OpenAI
        - Anthropic
        - Google
        - LiteLLM
        - Ollama

    Supported standard tools:
        - python_repl
        - serper.dev search
    """

    name = "langchain"

    def __init__(self, model_storage: ModelStorage, engine_storage: HandlerStorage, **kwargs):
        super().__init__(model_storage, engine_storage, **kwargs)
        # if True, the target column name does not have to be specified at creation time.
        self.generative = True
        self.default_agent_tools = DEFAULT_AGENT_TOOLS

    def _create_embeddings_model(self, args: Dict):
        return construct_model_from_args(args)

    def _handle_parsing_errors(self, error: Exception) -> str:
        response = str(error)
        for p in _PARSING_ERROR_PREFIXES:
            if response.startswith(p):
                # As a somewhat dirty workaround, we accept the output formatted incorrectly and use it as a response.
                #
                # Ideally, in the future, we would write a parser that is more robust and flexible than the one Langchain uses.
                # Response is wrapped in ``
                logger.info("Handling parsing error, salvaging response...")
                response_output = response.split("`")
                if len(response_output) >= 2:
                    response = response_output[-2]

                # Wrap response in Langchain conversational react format.
                langchain_react_formatted_response = f"""Thought: Do I need to use a tool? No
AI: {response}"""
                return langchain_react_formatted_response
        return f"Agent failed with error:\n{str(error)}..."

    def create(self, target: str, args: Dict = None, **kwargs):
        self.default_agent_tools = args.get("tools", self.default_agent_tools)

        args = args["using"]
        args["target"] = target
        args["model_name"] = args.get("model_name", DEFAULT_MODEL_NAME)
        args["provider"] = args.get("provider", get_llm_provider(args))
        args["embedding_model_provider"] = args.get("embedding_model", get_embedding_model_provider(args))
        if args.get("mode") == "retrieval":
            # use default prompt template for retrieval i.e. RAG if not provided
            if "prompt_template" not in args:
                args["prompt_template"] = DEFAULT_RAG_PROMPT_TEMPLATE

        self.model_storage.json_set("args", args)

    @staticmethod
    def create_validation(_, args: Dict = None, **kwargs):
        if "using" not in args:
            raise Exception("LangChain engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args["using"]
        if "prompt_template" not in args:
            if not args.get("mode") == "retrieval":
                raise ValueError("Please provide a `prompt_template` for this engine.")

    def predict(self, df: pd.DataFrame, args: Dict = None) -> pd.DataFrame:
        """
        Dispatch is performed depending on the underlying model type. Currently, only the default text completion
        is supported.
        """
        pred_args = args["predict_params"] if args else {}
        args = self.model_storage.json_get("args")
        if "prompt_template" not in args and "prompt_template" not in pred_args:
            raise ValueError("This model expects a `prompt_template`, please provide one.")
        # Back compatibility for old models
        args["provider"] = args.get("provider", get_llm_provider(args))
        args["embedding_model_provider"] = args.get("embedding_model", get_embedding_model_provider(args))

        df = df.reset_index(drop=True)

        if pred_args.get("mode") == "chat_model":
            return self.call_llm(df, args, pred_args)

        agent = self.create_agent(df, args, pred_args)
        # Use last message as prompt, remove other questions.
        user_column = args.get("user_column", USER_COLUMN)
        if user_column not in df.columns:
            raise Exception(
                f"Expected user input in column `{user_column}`, which is not found in the input data. Either provide the column, or redefine the expected column at model creation (`USING user_column = 'value'`)"
            )  # noqa
        df.iloc[:-1, df.columns.get_loc(user_column)] = None
        return self.run_agent(df, agent, args, pred_args)

    def call_llm(self, df, args=None, pred_args=None):
        llm = create_chat_model({**args, **pred_args})

        user_column = args.get("user_column", USER_COLUMN)
        assistant_column = args.get("assistant_column", ASSISTANT_COLUMN)

        question = df[user_column].iloc[-1]
        resp = llm([HumanMessage(question)], stop=["\nObservation:", "\n\tObservation:"])

        return pd.DataFrame([resp.content], columns=[assistant_column])

    def create_agent(self, df: pd.DataFrame, args: Dict = None, pred_args: Dict = None) -> AgentExecutor:
        pred_args = pred_args if pred_args else {}

        # Set up tools.
        model_kwargs = get_chat_model_params({**args, **pred_args})
        llm = create_chat_model({**args, **pred_args})

        tools = setup_tools(llm, model_kwargs, pred_args, self.default_agent_tools)

        # Prefer prediction prompt template over original if provided.
        prompt_template = pred_args.get("prompt_template", args["prompt_template"])
        if "context" in pred_args:
            prompt_template += "\n\n" + "Useful information:\n" + pred_args["context"] + "\n"

        # Set up memory.
        memory = ConversationSummaryBufferMemory(
            llm=llm, max_token_limit=model_kwargs.get("max_tokens", DEFAULT_MAX_TOKENS), memory_key="chat_history"
        )
        memory.chat_memory.messages.insert(0, SystemMessage(content=prompt_template))
        # User - Assistant conversation. All except the last message.
        user_column = args.get("user_column", USER_COLUMN)
        assistant_column = args.get("assistant_column", ASSISTANT_COLUMN)
        for row in df[:-1].to_dict("records"):
            question = row[user_column]
            answer = row[assistant_column]
            if question:
                memory.chat_memory.add_user_message(question)
            if answer:
                memory.chat_memory.add_ai_message(answer)

        agent_type = args.get("agent_type", DEFAULT_AGENT_TYPE)
        agent_executor = initialize_agent(
            tools,
            llm,
            agent=agent_type,
            # Use custom output parser to handle flaky LLMs that don't ALWAYS conform to output format.
            agent_kwargs={"output_parser": SafeOutputParser()},
            # Calls the agent’s LLM Chain one final time to generate a final answer based on the previous steps
            early_stopping_method="generate",
            handle_parsing_errors=self._handle_parsing_errors,
            # Timeout per agent invocation.
            max_execution_time=pred_args.get(
                "timeout_seconds", args.get("timeout_seconds", DEFAULT_AGENT_TIMEOUT_SECONDS)
            ),
            max_iterations=pred_args.get("max_iterations", args.get("max_iterations", DEFAULT_MAX_ITERATIONS)),
            memory=memory,
            verbose=pred_args.get("verbose", args.get("verbose", True)),
        )
        return agent_executor

    def run_agent(self, df: pd.DataFrame, agent: AgentExecutor, args: Dict, pred_args: Dict) -> pd.DataFrame:
        # Prefer prediction time prompt template, if available.
        base_template = pred_args.get("prompt_template", args["prompt_template"])

        input_variables = []
        matches = list(re.finditer("{{(.*?)}}", base_template))

        for m in matches:
            input_variables.append(m[0].replace("{", "").replace("}", ""))
        empty_prompt_ids = np.where(df[input_variables].isna().all(axis=1).values)[0]

        base_template = base_template.replace("{{", "{").replace("}}", "}")
        prompts = []

        user_column = args.get("user_column", USER_COLUMN)
        for i, row in df.iterrows():
            if i not in empty_prompt_ids:
                prompt = PromptTemplate(input_variables=input_variables, template=base_template)
                kwargs = {}
                for col in input_variables:
                    kwargs[col] = row[col] if row[col] is not None else ""  # add empty quote if data is missing
                prompts.append(prompt.format(**kwargs))
            elif row.get(user_column):
                # Just add prompt
                prompts.append(row[user_column])

        def _invoke_agent_executor_with_prompt(agent_executor, prompt):
            if not prompt:
                return ""
            try:
                answer = agent_executor.invoke(prompt)
            except Exception as e:
                answer = str(e)
                if not answer.startswith("Could not parse LLM output: `"):
                    raise e
                answer = {"output": answer.removeprefix("Could not parse LLM output: `").removesuffix("`")}

            if "output" not in answer:
                # This should never happen unless Langchain changes invoke output format, but just in case.
                return agent_executor.run(prompt)
            return answer["output"]

        completions = []
        # max_workers defaults to number of processors on the machine multiplied by 5.
        # https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
        max_workers = args.get("max_workers", None)
        agent_timeout_seconds = args.get("timeout", DEFAULT_AGENT_TIMEOUT_SECONDS)
        executor = ContextThreadPoolExecutor(max_workers=max_workers)
        futures = [executor.submit(_invoke_agent_executor_with_prompt, agent, prompt) for prompt in prompts]
        try:
            for future in as_completed(futures, timeout=agent_timeout_seconds):
                completions.append(future.result())
        except TimeoutError:
            completions.append(
                f"I'm sorry! I couldn't generate a response within the allotted time ({agent_timeout_seconds} seconds). "
                "If you need more time for processing, you can adjust the timeout settings. "
                "Please refer to the documentation for instructions on how to change the timeout value. "
                "Feel free to try your request again."
            )
        # Can't use ThreadPoolExecutor as context manager since we need wait=False.
        executor.shutdown(wait=False)

        # Add null completion for empty prompts
        for i in sorted(empty_prompt_ids)[:-1]:
            completions.insert(i, None)

        pred_df = pd.DataFrame(completions, columns=[args["target"]])

        return pred_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        tables = ["info"]
        return pd.DataFrame(tables, columns=["tables"])

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        raise NotImplementedError("Fine-tuning is not supported for LangChain models")



"""remove streams

Revision ID: 6d748f2c7b0b
Revises: b8be148dbc85
Create Date: 2023-05-24 15:22:11.969791

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa


# revision identifiers, used by Alembic.
revision = '6d748f2c7b0b'
down_revision = 'b8be148dbc85'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('stream')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'stream',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('name', sa.VARCHAR(), nullable=False),
        sa.Column('stream_in', sa.VARCHAR(), nullable=False),
        sa.Column('stream_out', sa.VARCHAR(), nullable=False),
        sa.Column('anomaly_stream', sa.VARCHAR(), nullable=True),
        sa.Column('integration', sa.VARCHAR(), nullable=True),
        sa.Column('predictor', sa.VARCHAR(), nullable=False),
        sa.Column('company_id', sa.INTEGER(), nullable=True),
        sa.Column('updated_at', sa.DATETIME(), nullable=True),
        sa.Column('created_at', sa.DATETIME(), nullable=True),
        sa.Column('type', sa.VARCHAR(), nullable=True),
        sa.Column('connection_info', sa.VARCHAR(), nullable=True),
        sa.Column('learning_params', sa.VARCHAR(), nullable=True),
        sa.Column('learning_threshold', sa.INTEGER(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###



from sqlalchemy import func
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.query_context.context_controller import query_context_controller
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError
from mindsdb.interfaces.model.functions import get_project_record, get_project_records


class ViewController:
    def add(self, name, query, project_name):
        name = name.lower()
        from mindsdb.interfaces.database.database import DatabaseController

        database_controller = DatabaseController()
        project_databases_dict = database_controller.get_dict(filter_type="project")

        if project_name not in project_databases_dict:
            raise EntityNotExistsError("Can not find project", project_name)

        project_id = project_databases_dict[project_name]["id"]
        view_record = (
            db.session.query(db.View.id)
            .filter(
                func.lower(db.View.name) == name, db.View.company_id == ctx.company_id, db.View.project_id == project_id
            )
            .first()
        )
        if view_record is not None:
            raise EntityExistsError("View already exists", name)

        view_record = db.View(name=name, company_id=ctx.company_id, query=query, project_id=project_id)
        db.session.add(view_record)
        db.session.commit()

    def update(self, name: str, query: str, project_name: str, strict_case: bool = False):
        """Update the SQL query of an existing view in the specified project.

        Args:
            name (str): The name of the view to update.
            query (str): The new SQL query for the view.
            project_name (str): The name of the project containing the view.
            strict_case (bool, optional): If True, the view name is case-sensitive. If False, the name comparison is case-insensitive. Defaults to False.

        Raises:
            EntityNotExistsError: If the view with the specified name does not exist in the given project.

        Returns:
            None
        """
        project_record = get_project_record(project_name)

        q = db.session.query(db.View).filter(
            db.View.company_id == ctx.company_id, db.View.project_id == project_record.id
        )
        if strict_case:
            q = q.filter(db.View.name == name)
        else:
            q = q.filter(func.lower(db.View.name) == func.lower(name))

        rec = q.first()
        if rec is None:
            raise EntityNotExistsError("View not found", name)
        rec.query = query
        db.session.commit()

    def delete(self, name: str, project_name: str, strict_case: bool = False) -> None:
        """Remove a view with the specified name from the given project.

        Args:
            name (str): The name of the view to remove.
            project_name (str): The name of the project containing the view.
            strict_case (bool, optional): If True, the view name is case-sensitive. Defaults to False.

        Raises:
            EntityNotExistsError: If the view does not exist.

        Returns:
            None
        """
        project_record = get_project_record(project_name)

        query = db.session.query(db.View).filter(
            db.View.company_id == ctx.company_id, db.View.project_id == project_record.id
        )
        if strict_case:
            query = query.filter(db.View.name == name)
        else:
            query = query.filter(func.lower(db.View.name) == func.lower(name))

        record = query.first()
        if record is None:
            raise EntityNotExistsError("View not found", name)
        db.session.delete(record)
        db.session.commit()

        query_context_controller.drop_query_context("view", record.id)

    def list(self, project_name):
        project_names = {}
        for project in get_project_records():
            if project_name is not None and project.name != project_name:
                continue
            project_names[project.id] = project.name

        query = db.session.query(db.View).filter(
            db.View.company_id == ctx.company_id, db.View.project_id.in_(list(project_names.keys()))
        )

        data = []

        for record in query:
            data.append(
                {
                    "id": record.id,
                    "name": record.name,
                    "project": project_names[record.project_id],
                    "query": record.query,
                }
            )

        return data

    def _get_view_record_data(self, record):
        return {"id": record.id, "name": record.name, "query": record.query}

    def get(self, id=None, name=None, project_name=None):
        project_record = get_project_record(project_name)

        if id is not None:
            records = (
                db.session.query(db.View)
                .filter_by(id=id, project_id=project_record.id, company_id=ctx.company_id)
                .all()
            )
        elif name is not None:
            records = (
                db.session.query(db.View)
                .filter(
                    func.lower(db.View.name) == name.lower(),
                    db.View.project_id == project_record.id,
                    db.View.company_id == ctx.company_id,
                )
                .all()
            )
        if len(records) == 0:
            if name is None:
                name = f"id={id}"
            raise EntityNotExistsError("Can't find view", f"{project_name}.{name}")
        elif len(records) > 1:
            raise Exception(f"There are multiple views with name/id: {name}/{id}")
        record = records[0]
        return self._get_view_record_data(record)



from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.utilities import log

from mindsdb_sql_parser import ast
from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
    INSERTQueryParser,
)

import pandas as pd
import re
from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.formatters import JSONFormatter

logger = log.getLogger(__name__)


class YoutubeCommentsTable(APITable):
    """Youtube List Comments  by video id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the youtube  "commentThreads()" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            youtube "commentThreads()" matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(query, "comments", self.get_columns())

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        channel_id, video_id = None, None
        for a_where in where_conditions:
            if a_where[1] == "video_id":
                if a_where[0] != "=":
                    raise NotImplementedError("Only '=' operator is supported for video_id column.")
                else:
                    video_id = a_where[2]
            elif a_where[1] == "channel_id":
                if a_where[0] != "=":
                    raise NotImplementedError("Only '=' operator is supported for channel_id column.")
                else:
                    channel_id = a_where[2]

        if not video_id and not channel_id:
            raise ValueError("Either video_id or channel_id has to be present in where clause.")

        comments_df = self.get_comments(video_id=video_id, channel_id=channel_id)

        select_statement_executor = SELECTQueryExecutor(
            comments_df,
            selected_columns,
            [
                where_condition
                for where_condition in where_conditions
                if where_condition[1] not in ["video_id", "channel_id"]
            ],
            order_by_conditions,
            result_limit if query.limit else None,
        )

        comments_df = select_statement_executor.execute_query()

        return comments_df

    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the YouTube POST /commentThreads API endpoint.

        Parameters
        ----------
        query : ast.Insert
            Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        insert_query_parser = INSERTQueryParser(query, self.get_columns())

        values_to_insert = insert_query_parser.parse_query()

        for value in values_to_insert:
            if not value.get("comment_id"):
                if not value.get("comment"):
                    raise ValueError("comment is mandatory for inserting a top-level comment.")
                else:
                    self.insert_comment(video_id=value["video_id"], text=value["comment"])

            else:
                if not value.get("reply"):
                    raise ValueError("reply is mandatory for inserting a reply.")
                else:
                    self.insert_comment(comment_id=value["comment_id"], text=value["reply"])

    def insert_comment(self, text, video_id: str = None, comment_id: str = None):
        # if comment_id is provided, define the request body for a reply and insert it
        if comment_id:
            request_body = {"snippet": {"parentId": comment_id, "textOriginal": text}}

            self.handler.connect().comments().insert(part="snippet", body=request_body).execute()

        # else if video_id is provided, define the request body for a top-level comment and insert it
        elif video_id:
            request_body = {"snippet": {"topLevelComment": {"snippet": {"videoId": video_id, "textOriginal": text}}}}

            self.handler.connect().commentThreads().insert(part="snippet", body=request_body).execute()

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "comment_id",
            "channel_id",
            "video_id",
            "user_id",
            "display_name",
            "comment",
            "published_at",
            "updated_at",
            "reply_user_id",
            "reply_author",
            "reply",
        ]

    def get_comments(self, video_id: str, channel_id: str):
        """Pulls all the records from the given youtube api end point and returns it select()

        Returns
        -------
        pd.DataFrame of all the records of the "commentThreads()" API end point
        """

        if video_id and channel_id:
            channel_id = None

        resource = (
            self.handler.connect()
            .commentThreads()
            .list(
                part="snippet, replies",
                videoId=video_id,
                allThreadsRelatedToChannelId=channel_id,
                textFormat="plainText",
            )
        )

        data = []
        while resource:
            comments = resource.execute()

            for comment in comments["items"]:
                replies = []
                if "replies" in comment:
                    for reply in comment["replies"]["comments"]:
                        replies.append(
                            {
                                "reply_author": reply["snippet"]["authorDisplayName"],
                                "user_id": reply["snippet"]["authorChannelId"]["value"],
                                "reply": reply["snippet"]["textOriginal"],
                            }
                        )
                else:
                    replies.append(
                        {
                            "reply_author": None,
                            "user_id": None,
                            "reply": None,
                        }
                    )

                data.append(
                    {
                        "channel_id": comment["snippet"]["channelId"],
                        "video_id": comment["snippet"]["videoId"],
                        "user_id": comment["snippet"]["topLevelComment"]["snippet"]["authorChannelId"]["value"],
                        "comment_id": comment["snippet"]["topLevelComment"]["id"],
                        "display_name": comment["snippet"]["topLevelComment"]["snippet"]["authorDisplayName"],
                        "comment": comment["snippet"]["topLevelComment"]["snippet"]["textDisplay"],
                        "published_at": comment["snippet"]["topLevelComment"]["snippet"]["publishedAt"],
                        "updated_at": comment["snippet"]["topLevelComment"]["snippet"]["updatedAt"],
                        "replies": replies,
                    }
                )

            if "nextPageToken" in comments:
                resource = (
                    self.handler.connect()
                    .commentThreads()
                    .list(
                        part="snippet, replies",
                        videoId=video_id,
                        allThreadsRelatedToChannelId=channel_id,
                        textFormat="plainText",
                        pageToken=comments["nextPageToken"],
                    )
                )
            else:
                break

        youtube_comments_df = pd.json_normalize(
            data,
            "replies",
            [
                "comment_id",
                "channel_id",
                "video_id",
                "user_id",
                "display_name",
                "comment",
                "published_at",
                "updated_at",
            ],
            record_prefix="replies.",
        )
        youtube_comments_df = youtube_comments_df.rename(
            columns={
                "replies.user_id": "reply_user_id",
                "replies.reply_author": "reply_author",
                "replies.reply": "reply",
            }
        )

        # check if DataFrame is empty
        if youtube_comments_df.empty:
            return youtube_comments_df
        else:
            return youtube_comments_df[
                [
                    "comment_id",
                    "channel_id",
                    "video_id",
                    "user_id",
                    "display_name",
                    "comment",
                    "published_at",
                    "updated_at",
                    "reply_user_id",
                    "reply_author",
                    "reply",
                ]
            ]


class YoutubeChannelsTable(APITable):
    """Youtube Channel Info  by channel id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        select_statement_parser = SELECTQueryParser(query, "channel", self.get_columns())

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        channel_id = None
        for op, arg1, arg2 in where_conditions:
            if arg1 == "channel_id":
                if op == "=":
                    channel_id = arg2
                    break
                else:
                    raise NotImplementedError("Only '=' operator is supported for channel_id column.")

        if not channel_id:
            raise NotImplementedError("channel_id has to be present in where clause.")

        channel_df = self.get_channel_details(channel_id)

        select_statement_executor = SELECTQueryExecutor(
            channel_df,
            selected_columns,
            [where_condition for where_condition in where_conditions if where_condition[1] == "channel_id"],
            order_by_conditions,
            result_limit if query.limit else None,
        )

        channel_df = select_statement_executor.execute_query()

        return channel_df

    def get_channel_details(self, channel_id):
        details = (
            self.handler.connect().channels().list(part="statistics,snippet,contentDetails", id=channel_id).execute()
        )
        snippet = details["items"][0]["snippet"]
        statistics = details["items"][0]["statistics"]
        data = {
            "country": snippet["country"],
            "description": snippet["description"],
            "creation_date": snippet["publishedAt"],
            "title": snippet["title"],
            "subscriber_count": statistics["subscriberCount"],
            "video_count": statistics["videoCount"],
            "view_count": statistics["viewCount"],
            "channel_id": channel_id,
        }
        return pd.json_normalize(data)

    def get_columns(self) -> List[str]:
        return [
            "country",
            "description",
            "creation_date",
            "title",
            "subscriber_count",
            "video_count",
            "view_count",
            "channel_id",
        ]


class YoutubeVideosTable(APITable):
    """Youtube Video info  by video id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        select_statement_parser = SELECTQueryParser(query, "video", self.get_columns())

        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        video_id, channel_id, search_query = None, None, None
        for op, arg1, arg2 in where_conditions:
            if arg1 == "video_id":
                if op == "=":
                    video_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for video_id column.")

            elif arg1 == "channel_id":
                if op == "=":
                    channel_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for channel_id column.")

            elif arg1 == "query":
                if op == "=":
                    search_query = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for query column.")

        if not video_id and not channel_id and not search_query:
            raise ValueError("At least one of video_id, channel_id, or query must be present in the WHERE clause.")

        if video_id:
            video_df = self.get_videos_by_video_ids([video_id])
        elif channel_id and search_query:
            video_df = self.get_videos_by_search_query_in_channel(search_query, channel_id, result_limit)
        elif channel_id:
            video_df = self.get_videos_by_channel_id(channel_id, result_limit)
        else:
            video_df = self.get_videos_by_search_query(search_query, result_limit)

        select_statement_executor = SELECTQueryExecutor(
            video_df,
            selected_columns,
            [
                where_condition
                for where_condition in where_conditions
                if where_condition[1] not in ["video_id", "channel_id", "query"]
            ],
            order_by_conditions,
            result_limit if query.limit else None,
        )

        video_df = select_statement_executor.execute_query()

        return video_df

    def get_videos_by_search_query(self, search_query, limit=10):
        video_ids = []
        resource = (
            self.handler.connect()
            .search()
            .list(part="snippet", q=search_query, type="video", maxResults=min(50, limit))
        )
        total_fetched = 0

        while resource and total_fetched < limit:
            response = resource.execute()
            for item in response["items"]:
                video_ids.append(item["id"]["videoId"])
                total_fetched += 1
                if total_fetched >= limit:
                    break

            if "nextPageToken" in response and total_fetched < limit:
                resource = (
                    self.handler.connect()
                    .search()
                    .list(
                        part="snippet",
                        q=search_query,
                        type="video",
                        maxResults=min(50, limit - total_fetched),
                        pageToken=response["nextPageToken"],
                    )
                )
            else:
                break

        return self.get_videos_by_video_ids(video_ids)

    def get_videos_by_search_query_in_channel(self, search_query, channel_id, limit=10):
        """Search for videos within a specific channel"""
        video_ids = []
        resource = (
            self.handler.connect()
            .search()
            .list(part="snippet", q=search_query, channelId=channel_id, type="video", maxResults=min(50, limit))
        )
        total_fetched = 0

        while resource and total_fetched < limit:
            response = resource.execute()
            for item in response["items"]:
                video_ids.append(item["id"]["videoId"])
                total_fetched += 1
                if total_fetched >= limit:
                    break

            if "nextPageToken" in response and total_fetched < limit:
                resource = (
                    self.handler.connect()
                    .search()
                    .list(
                        part="snippet",
                        q=search_query,
                        channelId=channel_id,
                        type="video",
                        maxResults=min(50, limit - total_fetched),
                        pageToken=response["nextPageToken"],
                    )
                )
            else:
                break

        return self.get_videos_by_video_ids(video_ids)

    def get_videos_by_channel_id(self, channel_id, limit=10):
        video_ids = []
        resource = (
            self.handler.connect()
            .search()
            .list(part="snippet", channelId=channel_id, type="video", maxResults=min(50, limit))
        )
        total_fetched = 0
        while resource and total_fetched < limit:
            response = resource.execute()
            for item in response["items"]:
                video_ids.append(item["id"]["videoId"])
                total_fetched += 1
                if total_fetched >= limit:
                    break
            if "nextPageToken" in response and total_fetched < limit:
                resource = (
                    self.handler.connect()
                    .search()
                    .list(
                        part="snippet",
                        channelId=channel_id,
                        type="video",
                        maxResults=min(50, limit - total_fetched),
                        pageToken=response["nextPageToken"],
                    )
                )
            else:
                break

        return self.get_videos_by_video_ids(video_ids)

    def get_videos_by_video_ids(self, video_ids):
        data = []

        if not isinstance(video_ids, list):
            logger.error(f"video_ids must be a list. Received {type(video_ids)} instead.")
            return pd.DataFrame()

        # loop over 50 video ids at a time
        # an invalid request error is caused otherwise
        for i in range(0, len(video_ids), 50):
            resource = (
                self.handler.connect()
                .videos()
                .list(part="statistics,snippet,contentDetails", id=",".join(video_ids[i : i + 50]))
                .execute()
            )

            for item in resource["items"]:
                data.append(
                    {
                        "channel_id": item["snippet"]["channelId"],
                        "channel_title": item["snippet"]["channelTitle"],
                        "comment_count": item["statistics"]["commentCount"],
                        "description": item["snippet"]["description"],
                        "like_count": item["statistics"]["likeCount"],
                        "publish_time": item["snippet"]["publishedAt"],
                        "title": item["snippet"]["title"],
                        "transcript": self.get_captions_by_video_id(item["id"]),
                        "video_id": item["id"],
                        "view_count": item["statistics"]["viewCount"],
                        "duration_str": self.parse_duration(item["id"], item["contentDetails"]["duration"]),
                    }
                )

        return pd.json_normalize(data)

    def get_captions_by_video_id(self, video_id):
        try:
            transcript_response = YouTubeTranscriptApi.get_transcript(video_id, preserve_formatting=True)
            json_formatted_transcript = JSONFormatter().format_transcript(transcript_response, indent=2)
            return json_formatted_transcript

        except Exception as e:
            (logger.error(f"Encountered an error while fetching transcripts for video ${video_id}: ${e}"),)
            return "Transcript not available for this video"

    def parse_duration(self, video_id, duration):
        try:
            parsed_duration = re.search(r"PT(\d+H)?(\d+M)?(\d+S)", duration).groups()
            duration_str = ""
            for d in parsed_duration:
                if d:
                    duration_str += f"{d[:-1]}:"

            return duration_str.strip(":")
        except Exception as e:
            (logger.error(f"Encountered an error while parsing duration for video ${video_id}: ${e}"),)
            return "Duration not available for this video"

    def get_columns(self) -> List[str]:
        return [
            "channel_id",
            "channel_title",
            "title",
            "description",
            "publish_time",
            "comment_count",
            "like_count",
            "view_count",
            "video_id",
            "duration_str",
            "transcript",
        ]



from urllib.parse import urlparse
import socket
import ipaddress


def is_private_url(url: str):
    """
    Raises exception if url is private

    :param url: url to check
    """

    hostname = urlparse(url).hostname
    if not hostname:
        # Unable find hostname in url
        return True
    ip = socket.gethostbyname(hostname)
    return ipaddress.ip_address(ip).is_private


def clear_filename(filename: str) -> str:
    """
    Removes path symbols from filename which could be used for path injection
    :param filename: input filename
    :return: output filename
    """

    if not filename:
        return filename
    badchars = '\\/:*?"<>|'
    for c in badchars:
        filename = filename.replace(c, "")
    return filename


def _split_url(url: str) -> tuple[str, str]:
    """
    Splits the URL into scheme and netloc.

    Args:
        url (str): The URL to split.

    Returns:
        tuple[str, str]: The scheme and netloc of the URL.

    Raises:
        ValueError: If the URL does not include protocol and host name.
    """
    parsed_url = urlparse(url)
    if not (parsed_url.scheme and parsed_url.netloc):
        raise ValueError(f"URL must include protocol and host name: {url}")
    return parsed_url.scheme.lower(), parsed_url.netloc.lower()


def validate_urls(urls: str | list[str], allowed_urls: list[str], disallowed_urls: list[str] | None = None) -> bool:
    """
    Checks if the provided URL(s) is/are from an allowed host.

    This function parses the URL(s) and checks the origin (scheme + netloc)
    against a list of allowed hosts.

    Examples:
        validate_urls("http://site.com/file", ["site.com"]) -> Exception
        validate_urls("https://site.com/file", ["https://site.com"]) -> True
        validate_urls("http://site.com/file", ["https://site.com"]) -> False
        validate_urls("https://site.com/file", ["https://example.com"]) -> False
        validate_urls("site.com/file", ["https://site.com"]) -> Exception

    Args:
        urls (str | list[str]): The URL(s) to check. Can be a single URL (str) or a list of URLs (list).
        allowed_urls (list[str]): The list of allowed URLs.
        disallowed_urls (list[str]): The list of disallowed URLs. If provided, the function
                                     will return False if the URL is in the disallowed list.

    Returns:
        bool: True if the URL(s) is/are from an allowed host and not in the disallowed list, False otherwise.
    """
    if disallowed_urls is None:
        disallowed_urls = []

    allowed_origins = [_split_url(url) for url in allowed_urls]
    disallowed_origins = [_split_url(url) for url in disallowed_urls]

    if isinstance(urls, str):
        urls = [urls]

    if allowed_origins:
        for url in urls:
            if _split_url(url) not in allowed_origins:
                return False

    if disallowed_origins:
        for url in urls:
            if _split_url(url) in disallowed_origins:
                return False

    return True



import unittest
from mindsdb.integrations.handlers.derby_handler.derby_handler import DerbyHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class DerbyHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "port": "1527",
                "database": "seconddb",
            }
        }
        cls.handler = DerbyHandler('test_derby_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        self.handler.check_connection()

    def test_2_create(self):
        res = self.handler.query('CREATE TABLE TESTTABLEX (ID INT PRIMARY KEY, NAME VARCHAR(14))')
        assert res.type is RESPONSE_TYPE.OK

    def test_3_insert(self):
        res = self.handler.query("INSERT INTO TESTTABLEX VALUES (100,'ONE HUNDRED'),(200,'TWO HUNDRED'),(300,'THREE HUNDRED')")
        assert res.type is RESPONSE_TYPE.OK

    def test_4_select(self):
        res = self.handler.query('SELECT * FROM TESTTABLEX')
        assert res.type is RESPONSE_TYPE.TABLE

    def test_5_get_tables(self):
        res = self.handler.get_tables()
        assert res.type is RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        res = self.handler.get_columns("TESTTABLEX")
        assert res.type is RESPONSE_TYPE.TABLE


if __name__ == '__main__':
    unittest.main()



from mindsdb_sql_parser.ast import Identifier

from mindsdb_sql_parser import parse_sql, ParsingException

from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.config import config

from mindsdb.api.executor.controllers.session_controller import SessionController


class TriggersController:
    OBJECT_TYPE = 'trigger'

    def add(self, name, project_name, table, query_str, columns=None):
        name = name.lower()

        if project_name is None:
            project_name = config.get('default_project')
        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        from mindsdb.api.executor.controllers.session_controller import SessionController
        session = SessionController()

        # check exists
        trigger = self.get_trigger_record(name, project_name)
        if trigger is not None:
            raise Exception(f'Trigger already exists: {name}')

        # check table
        if len(table.parts) < 2:
            raise Exception(f'Database or table not found: {table}')

        table_name = Identifier(parts=table.parts[1:]).to_string()
        db_name = table.parts[0]

        db_integration = session.integration_controller.get(db_name)
        db_handler = session.integration_controller.get_data_handler(db_name)

        if not hasattr(db_handler, 'subscribe'):
            raise Exception(f'Handler {db_integration["engine"]} does''t support subscription')

        df = db_handler.get_tables().data_frame
        column = 'table_name'
        if column not in df.columns:
            column = df.columns[0]
        tables = list(df[column])

        # check only if tables are visible
        if len(tables) > 0 and table_name not in tables:
            raise Exception(f'Table {table_name} not found in {db_name}')

        columns_str = None
        if columns is not None and len(columns) > 0:
            # join to string with delimiter
            columns_str = '|'.join([col.parts[-1] for col in columns])

        # check sql
        try:
            parse_sql(query_str)
        except ParsingException as e:
            raise ParsingException(f'Unable to parse: {query_str}: {e}')

        # create job record
        record = db.Triggers(
            name=name,
            project_id=project.id,

            database_id=db_integration['id'],
            table_name=table_name,
            query_str=query_str,
            columns=columns_str
        )
        db.session.add(record)
        db.session.flush()

        task_record = db.Tasks(
            company_id=ctx.company_id,
            user_class=ctx.user_class,

            object_type=self.OBJECT_TYPE,
            object_id=record.id,
        )
        db.session.add(task_record)
        db.session.commit()

    def delete(self, name, project_name):
        # check exists

        trigger = self.get_trigger_record(name, project_name)
        if trigger is None:
            raise Exception(f"Trigger doesn't exist: {name}")

        task = db.Tasks.query.filter(
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.object_id == trigger.id,
            db.Tasks.company_id == ctx.company_id,
        ).first()

        if task is not None:
            db.session.delete(task)

        db.session.delete(trigger)

        db.session.commit()

    def get_trigger_record(self, name, project_name):
        project_controller = ProjectController()
        project = project_controller.get(name=project_name)

        query = db.session.query(
            db.Triggers
        ).join(
            db.Tasks, db.Triggers.id == db.Tasks.object_id
        ).filter(
            db.Triggers.project_id == project.id,
            db.Triggers.name == name,
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.company_id == ctx.company_id,
        )
        return query.first()

    def get_list(self, project_name=None):
        session = SessionController()

        query = db.session.query(
            db.Tasks.object_id,
            db.Triggers.project_id,
            db.Triggers.name,
            db.Triggers.database_id,
            db.Triggers.table_name,
            db.Triggers.query_str,
            db.Tasks.last_error,
        )\
            .join(db.Triggers, db.Triggers.id == db.Tasks.object_id)\
            .filter(
                db.Tasks.object_type == self.OBJECT_TYPE,
                db.Tasks.company_id == ctx.company_id,
        )

        project_controller = ProjectController()
        if project_name is not None:
            project = project_controller.get(name=project_name)
            query = query.filter(db.Triggers.project_id == project.id)

        database_names = {
            i['id']: i['name']
            for i in session.database_controller.get_list()
        }

        project_names = {
            i.id: i.name
            for i in project_controller.get_list()
        }
        data = []
        for record in query:
            data.append({
                'id': record.object_id,
                'project': project_names[record.project_id],
                'name': record.name.lower(),
                'database': database_names.get(record.database_id, '?'),
                'table': record.table_name,
                'query': record.query_str,
                'last_error': record.last_error,
            })
        return data
